#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2008 Jonas Termeau - jonas.termeau **AT** gmail.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Thanks to:
#
# Bernard Gray - bernard.gray **AT** gmail.com (python helping)
# Jamie Heames (english translation issues)
# ~suv (bug report in v2.3)
# http://www.gutenberg.eu.org/publications/ (9x9 margins settings)
#
"""
This basic extension allows you to automatically draw guides in inkscape.
"""

from math import cos, sin, sqrt

import inkex
from inkex import Guide


class GuidesCreator(inkex.EffectExtension):
    """Create a set of guides based on the given options"""
    def add_arguments(self, pars):
        pars.add_argument("--tab", type=self.arg_method('generate'), default="regular_guides",\
            help="Type of guides to create.")
        pars.add_argument('--guides_preset', default='custom', help='Preset')
        pars.add_argument('--vertical_guides', type=int, default=3, help='Vertical guides')
        pars.add_argument('--horizontal_guides', type=int, default=3, help='Horizontal guides')
        pars.add_argument('--start_from_edges', type=inkex.Boolean, help='Start from edges')
        pars.add_argument('--ul', type=inkex.Boolean, default=True, help='Upper left corner')
        pars.add_argument('--ur', type=inkex.Boolean, default=True, help='Upper right corner')
        pars.add_argument('--ll', type=inkex.Boolean, default=True, help='Lower left corner')
        pars.add_argument('--lr', type=inkex.Boolean, default=True, help='Lower right corner')
        pars.add_argument('--margins_preset', default='custom', help='Margins preset')
        pars.add_argument('--vert', type=int, default=0, help='Vert subdivisions')
        pars.add_argument('--horz', type=int, default=0, help='Horz subdivisions')
        pars.add_argument('--header_margin', default=6, help='Header margin')
        pars.add_argument('--footer_margin', default=6, help='Footer margin')
        pars.add_argument('--left_margin', default=6, help='Left margin')
        pars.add_argument('--right_margin', default=6, help='Right margin')
        pars.add_argument('--delete', type=inkex.Boolean, help='Delete existing guides')

    def effect(self):
        # getting the width and height attributes of the canvas
        self.width = float(self.svg.width)
        self.height = float(self.svg.height)

        # getting edges coordinates
        self.h_orientation = '0,' + str(round(self.width, 4))
        self.v_orientation = str(round(self.height, 4)) + ',0'

        if self.options.delete:
            for guide in self.svg.namedview.get_guides():
                guide.delete()

        return self.options.tab()

    def generate_regular_guides(self):
        """Generate a regular set of guides"""
        preset = self.options.guides_preset
        from_edges = self.options.start_from_edges
        if preset == 'custom':
            h_division = self.options.horizontal_guides
            v_division = self.options.vertical_guides
            if from_edges:
                v_division = v_division or 1
                h_division = h_division or 1

            self.draw_guides(v_division, from_edges, vert=True)
            self.draw_guides(h_division, from_edges, vert=False)

        elif preset == 'golden':
            gold = (1 + sqrt(5)) / 2

            # horizontal golden guides
            position1 = '0,' + str(self.height / gold)
            position2 = '0,' + str(self.height - (self.height / gold))

            self.draw_guide(position1, self.h_orientation)
            self.draw_guide(position2, self.h_orientation)

            # vertical golden guides
            position1 = str(self.width / gold) + ',0'
            position2 = str(self.width - (self.width / gold)) + ',0'

            self.draw_guide(position1, self.v_orientation)
            self.draw_guide(position2, self.v_orientation)

            if from_edges:
                # horizontal borders
                self.draw_guide('0,' + str(self.height), self.h_orientation)
                self.draw_guide(str(self.height) + ',0', self.h_orientation)

                # vertical borders
                self.draw_guide('0,' + str(self.width), self.v_orientation)
                self.draw_guide(str(self.width) + ',0', self.v_orientation)

        elif ';' in preset:
            v_division = int(preset.split(';')[0])
            h_division = int(preset.split(';')[1])
            self.draw_guides(v_division, from_edges, vert=True)
            self.draw_guides(h_division, from_edges, vert=False)
        else:
            raise inkex.AbortExtension("Unknown guide guide preset: {}".format(preset))

    def generate_diagonal_guides(self):
        """Generate diagonal guides"""
        # Dimentions
        left, bottom = (0, 0)
        right, top = (self.width, self.height)

        # Diagonal angle
        angle = 45

        if self.options.ul:
            ul_corner = str(top) + ',' + str(left)
            from_ul_to_lr = str(cos(angle)) + ',' + str(cos(angle))
            self.draw_guide(ul_corner, from_ul_to_lr)

        if self.options.ur:
            ur_corner = str(right) + ',' + str(top)
            from_ur_to_ll = str(-sin(angle)) + ',' + str(sin(angle))
            self.draw_guide(ur_corner, from_ur_to_ll)

        if self.options.ll:
            ll_corner = str(bottom) + ',' + str(left)
            from_ll_to_ur = str(-cos(angle)) + ',' + str(cos(angle))
            self.draw_guide(ll_corner, from_ll_to_ur)

        if self.options.lr:
            lr_corner = str(bottom) + ',' + str(right)
            from_lr_to_ul = str(-sin(angle)) + ',' + str(-sin(angle))
            self.draw_guide(lr_corner, from_lr_to_ul)

    def generate_margins(self):
        """Generate margin guides"""
        header_margin = int(self.options.header_margin)
        footer_margin = int(self.options.footer_margin)
        left_margin = int(self.options.left_margin)
        right_margin = int(self.options.right_margin)
        h_subdiv = int(self.options.horz)
        v_subdiv = int(self.options.vert)

        if self.options.start_from_edges:
            # horizontal borders
            self.draw_guide('0,' + str(self.height), self.h_orientation)
            self.draw_guide(str(self.height) + ',0', self.h_orientation)

            # vertical borders
            self.draw_guide('0,' + str(self.width), self.v_orientation)
            self.draw_guide(str(self.width) + ',0', self.v_orientation)

        if self.options.margins_preset == 'custom':
            y_header = self.height
            y_footer = 0
            x_left = 0
            x_right = self.width

            if header_margin != 0:
                y_header = (self.height / header_margin) * (header_margin - 1)
                self.draw_guide('0,' + str(y_header), self.h_orientation)

            if footer_margin != 0:
                y_footer = self.height / footer_margin
                self.draw_guide('0,' + str(y_footer), self.h_orientation)

            if left_margin != 0:
                x_left = self.width / left_margin
                self.draw_guide(str(x_left) + ',0', self.v_orientation)

            if right_margin != 0:
                x_right = (self.width / right_margin) * (right_margin - 1)
                self.draw_guide(str(x_right) + ',0', self.v_orientation)

        elif self.options.margins_preset == 'book_left':
            # 1/9th header
            y_header = (self.height / 9) * 8
            self.draw_guide('0,' + str(y_header), self.h_orientation)

            # 2/9th footer
            y_footer = (self.height / 9) * 2
            self.draw_guide('0,' + str(y_footer), self.h_orientation)

            # 2/9th left margin
            x_left = (self.width / 9) * 2
            self.draw_guide(str(x_left) + ',0', self.v_orientation)

            # 1/9th right margin
            x_right = (self.width / 9) * 8
            self.draw_guide(str(x_right) + ',0', self.v_orientation)

        elif self.options.margins_preset == 'book_right':
            # 1/9th header
            y_header = (self.height / 9) * 8
            self.draw_guide('0,' + str(y_header), self.h_orientation)

            # 2/9th footer
            y_footer = (self.height / 9) * 2
            self.draw_guide('0,' + str(y_footer), self.h_orientation)

            # 2/9th left margin
            x_left = (self.width / 9)
            self.draw_guide(str(x_left) + ',0', self.v_orientation)

            # 1/9th right margin
            x_right = (self.width / 9) * 7
            self.draw_guide(str(x_right) + ',0', self.v_orientation)

        # setting up properties of the rectangle created between guides
        rectangle_height = y_header - y_footer
        rectangle_width = x_right - x_left

        if h_subdiv != 0:
            begin_from = y_footer
            # creating horizontal guides
            self._draw_guides(
                (rectangle_width, rectangle_height), h_subdiv,
                edges=0, shift=begin_from, vert=False)

        if v_subdiv != 0:
            begin_from = x_left
            # creating vertical guides
            self._draw_guides(
                (rectangle_width, rectangle_height), v_subdiv,
                edges=0, shift=begin_from, vert=True)

    def draw_guides(self, division, edges, vert=False):
        """Draw a vertical or horizontal lines"""
        return self._draw_guides((self.width, self.height), division, edges, vert=vert)

    def _draw_guides(self, vector, division, edges, shift=0, vert=False):
        if division <= 0:
            return

        # Vert controls both ort template and vector calculation
        ort = '{},0' if vert else '0,{}'
        var = int(bool(edges))
        for x in range(0, division - 1 + 2 * var):
            div = vector[not bool(vert)] / division
            position = str(round(div + (x - var) * div + shift, 4))
            orientation = str(round(vector[bool(vert)], 4))
            self.draw_guide(ort.format(position), ort.format(orientation))

    def draw_guide(self, position, orientation):
        """Create a guide directly into the namedview"""
        if isinstance(position, tuple):
            x, y = position
        elif isinstance(position, str):
            x, y = position.split(',')
        self.svg.namedview.add(Guide().move_to(float(x), float(y), orientation))

if __name__ == '__main__':
    GuidesCreator().run()



#!/usr/bin/env python
"""Extension to remove the blue colour from selected shapes"""

import inkex

class RemoveBlue(inkex.ColorExtension):
    """Remove blue color from selected objects"""
    def modify_color(self, name, color):
        return inkex.Color([color.red, color.green, 0])

if __name__ == '__main__':
    RemoveBlue().run()



#!/usr/bin/env python
"""Darken colours of selected objects"""

import inkex

class Darker(inkex.ColorExtension):
    """Make the colours darker"""
    def modify_color(self, name, color):
        factor = 0.9
        if color.space == 'hsl':
            color.lightness = int(round(max(color.lightness * factor, 0)))
        else:
            color.red = int(round(max(color.red * factor, 0)))
            color.green = int(round(max(color.green * factor, 0)))
            color.blue = int(round(max(color.blue * factor, 0)))
        return color

if __name__ == '__main__':
    Darker().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2007 John Beard john.j.beard@gmail.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This extension allows you to draw a polar grid in Inkscape.
There is a wide range of options including subdivision and labels.
"""

from math import cos, log, pi, sin

import inkex
from inkex import Group, Circle, TextElement

def draw_circle(r, cx, cy, width, fill, name, parent):
    """Draw an SVG circle"""
    circle = parent.add(Circle(cx=str(cx), cy=str(cy), r=str(r)))
    circle.style = {'stroke': '#000000', 'stroke-width': str(width), 'fill': fill}
    circle.label = name

def draw_line(x1, y1, x2, y2, width, name, parent):
    """Draw an SVG line"""
    line = parent.add(inkex.PathElement())
    line.style = {'stroke': '#000000', 'stroke-width': str(width), 'fill': 'none'}
    line.path = 'M {},{} L {},{}'.format(x1, y1, x2, y2)
    line.label = name

def draw_label(x, y, string, font_size, name, parent):
    """Draw a centered label"""
    label = parent.add(TextElement(x=str(x), y=str(y)))
    label.style = {'text-align': 'center', 'vertical-align': 'top',
                   'text-anchor': 'middle', 'font-size': str(font_size) + 'px',
                   'fill-opacity': '1.0', 'stroke': 'none',
                   'font-weight': 'normal', 'font-style': 'normal', 'fill': '#000000'}
    label.text = string
    label.label = name


class GridPolar(inkex.GenerateExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--r_divs", type=int, default=5, help="Circular Divisions")
        pars.add_argument("--dr", type=float, default=50, help="Circular Division Spacing")
        pars.add_argument("--r_subdivs", type=int, default=3, help="Subdivisions per Major div")
        pars.add_argument("--r_log", type=inkex.Boolean, default=False, help="Logarithmic subdiv")
        pars.add_argument("--r_divs_th", type=float, default=2, help="Major Line thickness")
        pars.add_argument("--r_subdivs_th", type=float, default=1, help="Minor Line thickness")
        pars.add_argument("--a_divs", type=int, default=24, help="Angle Divisions")
        pars.add_argument("--a_divs_cent", type=int, default=4, help="Angle Divisions at Centre")
        pars.add_argument("--a_subdivs", type=int, default=1, help="Angcular Subdivisions")
        pars.add_argument("--a_subdivs_cent", type=int, default=1, help="Angular Subdivisions end")
        pars.add_argument("--a_divs_th", type=float, default=2, help="Major Angular thickness")
        pars.add_argument("--a_subdivs_th", type=float, default=1, help="Minor Angular thickness")
        pars.add_argument("--c_dot_dia", type=float, default=5.0, help="Diameter of Centre Dot")
        pars.add_argument("--a_labels", default='deg', help="The kind of labels to apply")
        pars.add_argument("--a_label_size", type=int, default=18, help="Pixel size of the labels")
        pars.add_argument("--a_label_outset", type=float, default=24, help="Label Radial outset")

    def generate(self):
        self.options.dr = self.svg.unittouu(str(self.options.dr) + 'px')
        self.options.r_divs_th = self.svg.unittouu(str(self.options.r_divs_th) + 'px')
        self.options.r_subdivs_th = self.svg.unittouu(str(self.options.r_subdivs_th) + 'px')
        self.options.a_divs_th = self.svg.unittouu(str(self.options.a_divs_th) + 'px')
        self.options.a_subdivs_th = self.svg.unittouu(str(self.options.a_subdivs_th) + 'px')
        self.options.c_dot_dia = self.svg.unittouu(str(self.options.c_dot_dia) + 'px')
        self.options.a_label_size = self.svg.unittouu(str(self.options.a_label_size) + 'px')
        self.options.a_label_outset = self.svg.unittouu(str(self.options.a_label_outset) + 'px')

        # Embed grid in group
        grid = Group.new("GridPolar:R{0.r_divs}:A{0.a_divs}".format(self.options))

        (pos_x, pos_y) = self.svg.namedview.center
        grid.transform.add_translate(pos_x, pos_y)

        dr = self.options.dr  # Distance between neighbouring circles
        dtheta = 2 * pi / self.options.a_divs_cent  # Angular change between adjacent radial lines at centre
        rmax = self.options.r_divs * dr

        # Create SVG circles
        for i in range(1, self.options.r_divs + 1):
            draw_circle(i * dr, 0, 0,  # major div circles
                            self.options.r_divs_th, 'none',
                            'MajorDivCircle' + str(i) + ':R' + str(i * dr), grid)

            if self.options.r_log:  # logarithmic subdivisions
                for j in range(2, self.options.r_subdivs):
                    draw_circle(i * dr - (1 - log(j, self.options.r_subdivs)) * dr,  # minor div circles
                                    0, 0, self.options.r_subdivs_th, 'none',
                                    'MinorDivCircle' + str(i) + ':Log' + str(j), grid)
            else:  # linear subdivs
                for j in range(1, self.options.r_subdivs):
                    draw_circle(i * dr - j * dr / self.options.r_subdivs,  # minor div circles
                                    0, 0, self.options.r_subdivs_th, 'none',
                                    'MinorDivCircle' + str(i) + ':R' + str(i * dr), grid)

        if self.options.a_divs == self.options.a_divs_cent:  # the lines can go from the centre to the edge
            for i in range(0, self.options.a_divs):
                draw_line(0, 0, rmax * sin(i * dtheta), rmax * cos(i * dtheta),
                              self.options.a_divs_th, 'RadialGridline' + str(i), grid)

        else:  # we need separate lines
            for i in range(0, self.options.a_divs_cent):  # lines that go to the first circle
                draw_line(0, 0, dr * sin(i * dtheta), dr * cos(i * dtheta),
                              self.options.a_divs_th, 'RadialGridline' + str(i), grid)

            dtheta = 2 * pi / self.options.a_divs  # work out the angle change for outer lines

            for i in range(0, self.options.a_divs):  # lines that go from there to the edge
                draw_line(dr * sin(i * dtheta + pi / 2.0), dr * cos(i * dtheta + pi / 2.0),
                              rmax * sin(i * dtheta + pi / 2.0), rmax * cos(i * dtheta + pi / 2.0),
                              self.options.a_divs_th, 'RadialGridline' + str(i), grid)

        if self.options.a_subdivs > 1:  # draw angular subdivs
            for i in range(0, self.options.a_divs):  # for each major division
                for j in range(1, self.options.a_subdivs):  # draw the subdivisions
                    angle = i * dtheta - j * dtheta / self.options.a_subdivs + pi / 2.0  # the angle of the subdivion line
                    draw_line(dr * self.options.a_subdivs_cent * sin(angle),
                                  dr * self.options.a_subdivs_cent * cos(angle),
                                  rmax * sin(angle), rmax * cos(angle),
                                  self.options.a_subdivs_th, 'RadialMinorGridline' + str(i), grid)

        if self.options.c_dot_dia != 0:  # if a non-zero diameter, draw the centre dot
            draw_circle(self.options.c_dot_dia / 2.0,
                            0, 0, 0, '#000000', 'CentreDot', grid)

        if self.options.a_labels == 'deg':
            label_radius = rmax + self.options.a_label_outset  # radius of label centres
            label_size = self.options.a_label_size
            numeral_size = 0.73 * label_size  # numerals appear to be 0.73 the height of the nominal pixel size of the font in "Sans"

            for i in range(0, self.options.a_divs):  # self.options.a_divs): #radial line labels
                draw_label(sin(i * dtheta + pi / 2.0) * label_radius,  # 0 at the RHS, mathematical style
                                       cos(i * dtheta + pi / 2.0) * label_radius + numeral_size / 2.0,  # centre the text vertically
                                       str(i * 360 / self.options.a_divs),
                                       label_size, 'Label' + str(i), grid)

        return grid

if __name__ == '__main__':
    GridPolar().run()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Martin Owens <doctormo@gmail.com>
#                    Sergei Izmailov <sergei.a.izmailov@gmail.com>
#                    Thomas Holder <thomas.holder@schrodinger.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=arguments-differ
"""
Element interface for patterns, filters, gradients and path effects.
"""

from lxml import etree
from copy import deepcopy

from ..transforms import Transform
from ..tween import interpcoord, interp
from ..styles import Style

from ._utils import addNS
from ._base import BaseElement


from typing import overload, Iterable, List, Tuple, Union, Optional  # pylint: disable=unused-import


class Filter(BaseElement):
    """A filter (usually in defs)"""
    tag_name = 'filter'

    def add_primitive(self, fe_type, **args):
        """Create a filter primitive with the given arguments"""
        elem = etree.SubElement(self, addNS(fe_type, 'svg'))
        elem.update(**args)
        return elem

    class Primitive(BaseElement):
        pass

    class Blend(Primitive):
        tag_name = 'feBlend'

    class ColorMatrix(Primitive):
        tag_name = 'feColorMatrix'

    class ComponentTransfer(Primitive):
        tag_name = 'feComponentTransfer'

    class Composite(Primitive):
        tag_name = 'feComposite'

    class ConvolveMatrix(Primitive):
        tag_name = 'feConvolveMatrix'

    class DiffuseLighting(Primitive):
        tag_name = 'feDiffuseLighting'

    class DisplacementMap(Primitive):
        tag_name = 'feDisplacementMap'

    class Flood(Primitive):
        tag_name = 'feFlood'

    class GaussianBlur(Primitive):
        tag_name = 'feGaussianBlur'

    class Image(Primitive):
        tag_name = 'feImage'

    class Merge(Primitive):
        tag_name = 'feMerge'

    class Morphology(Primitive):
        tag_name = 'feMorphology'

    class Offset(Primitive):
        tag_name = 'feOffset'

    class SpecularLighting(Primitive):
        tag_name = 'feSpecularLighting'

    class Tile(Primitive):
        tag_name = 'feTile'

    class Turbulence(Primitive):
        tag_name = 'feTurbulence'


class Stop(BaseElement):
    tag_name = 'stop'

    @property
    def offset(self):
        # type: () -> float
        return self.get('offset')

    @offset.setter
    def offset(self, number):
        self.set('offset', number)

    def interpolate(self, other, fraction):
        newstop = Stop()
        newstop.style = self.style.interpolate(other.style, fraction)
        newstop.offset = interpcoord(float(self.offset), float(other.offset), fraction)
        return newstop


class Pattern(BaseElement):
    """Pattern element which is used in the def to control repeating fills"""
    tag_name = 'pattern'
    WRAPPED_ATTRS = BaseElement.WRAPPED_ATTRS + (('patternTransform', Transform),)


class Gradient(BaseElement):
    """A gradient instruction usually in the defs"""
    WRAPPED_ATTRS = BaseElement.WRAPPED_ATTRS + (('gradientTransform', Transform),)

    orientation_attributes = () # type: Tuple[str, ...]

    @property
    def stops(self):
        """Return an ordered list of own or linked stop nodes"""
        gradcolor = self.href if isinstance(self.href, LinearGradient) else self
        return sorted([child for child in gradcolor if isinstance(child, Stop)]
                      , key=lambda x: float(x.offset))

    @property
    def stop_offsets(self):
        # type: () -> List[float]
        """Return a list of own or linked stop offsets"""
        return [child.offset for child in self.stops]

    @property
    def stop_styles(self): # type: () -> List[Style]
        """Return a list of own or linked offset styles"""
        return [child.style for child in self.stops]

    def remove_orientation(self):
        """Remove all orientation attributes from this element"""
        for attr in self.orientation_attributes:
            self.pop(attr)

    def interpolate(self, other, fraction): # type: (LinearGradient, float) -> LinearGradient
        """Interpolate with another gradient."""
        if self.tag_name != other.tag_name:
            return self
        newgrad = self.copy()

        # interpolate transforms
        newtransform = self.gradientTransform.interpolate(other.gradientTransform, fraction)
        newgrad.gradientTransform = newtransform

        # interpolate orientation
        for attr in self.orientation_attributes:
            newattr = interpcoord(self.uutounit(self.get(attr)), self.uutounit(other.get(attr)), fraction)
            newgrad.set(attr, self.unittouu(newattr))

        # interpolate stops
        if self.href is not None and self.href is other.href:
            # both gradients link to the same stops
            pass
        else:
            # gradients might have different stops
            newoffsets = sorted(self.stop_offsets + other.stop_offsets[1:-1])
            func = lambda x,y,f: x.interpolate(y, f)
            sstops = interp(self.stop_offsets, list(self.stops), newoffsets, func)
            ostops = interp(other.stop_offsets, list(other.stops), newoffsets, func)
            newstops = [s1.interpolate(s2, fraction) for s1, s2 in zip(sstops, ostops)]
            newgrad.remove_all(Stop)
            newgrad.add(*newstops)
        return newgrad

    def stops_and_orientation(self):
        """Return a copy of all the stops in this gradient"""
        stops = self.copy()
        stops.remove_orientation()
        orientation = self.copy()
        orientation.remove_all(Stop)
        return stops, orientation


class LinearGradient(Gradient):
    tag_name = 'linearGradient'
    orientation_attributes = ('x1', 'y1', 'x2', 'y2')

    def apply_transform(self): # type: () -> None
       """Apply transform to orientation points and set it to identity."""
       trans = self.pop('gradientTransform')
       p1 = (self.uutounit(self.get('x1')), self.uutounit(self.get('y1')))
       p2 = (self.uutounit(self.get('x2')), self.uutounit(self.get('y2')))
       p1t = trans.apply_to_point(p1)
       p2t = trans.apply_to_point(p2)
       self.update(
            x1=self.unittouu(p1t[0]),
            y1=self.uutounit(p1t[1]),
            x2=self.uutounit(p2t[0]),
            y2=self.uutounit(p2t[1]))


class RadialGradient(Gradient):
    tag_name = 'radialGradient'
    orientation_attributes = ('cx', 'cy', 'fx', 'fy', 'r')

    def apply_transform(self): # type: () -> None
       """Apply transform to orientation points and set it to identity."""
       trans = self.pop('gradientTransform')
       p1 = (self.uutounit(self.get('cx')), self.uutounit(self.get('cy')))
       p2 = (self.uutounit(self.get('fx')), self.uutounit(self.get('fy')))
       p1t = trans.apply_to_point(p1)
       p2t = trans.apply_to_point(p2)
       self.update(
            cx=self.uutounit(p1t[0]),
            cy=self.uutounit(p1t[1]),
            fx=self.uutounit(p2t[0]),
            fy=self.uutounit(p2t[1]))

class PathEffect(BaseElement):
    """Inkscape LPE element"""
    tag_name = 'inkscape:path-effect'


class MeshGradient(Gradient):
    """Usable MeshGradient XML base class"""
    tag_name = 'meshgradient'

    @classmethod
    def new_mesh(cls, pos=None, rows=1, cols=1, autocollect=True):
        """Return skeleton of 1x1 meshgradient definition."""
        # initial point
        if pos is None or len(pos) != 2:
            pos = [0.0, 0.0]
        # create nested elements for rows x cols mesh
        meshgradient = cls()
        for _ in range(rows):
            meshrow = meshgradient.add(MeshRow())
            for _ in range(cols):
                meshrow.append(MeshPatch())
        # set meshgradient attributes
        meshgradient.set('gradientUnits', 'userSpaceOnUse')
        meshgradient.set('x', pos[0])
        meshgradient.set('y', pos[1])
        if autocollect:
            meshgradient.set('inkscape:collect', 'always')
        return meshgradient


class MeshRow(BaseElement):
    """Each row of a mesh gradient"""
    tag_name = 'meshrow'

class MeshPatch(BaseElement):
    """Each column or 'patch' in a mesh gradient"""
    tag_name = 'meshpatch'

    def stops(self, edges, colors):
        """Add or edit meshpatch stops with path and stop-color."""
        # iterate stops based on number of edges (path data)
        for i, edge in enumerate(edges):
            if i < len(self):
                stop = self[i]
            else:
                stop = self.add(Stop())

            # set edge path data
            stop.set('path', str(edge))
            # set stop color
            stop.style['stop-color'] = str(colors[i % 2])



# coding=utf-8
#
# Copyright (C) 2013 Public Domain
#               2018 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# standard libraries
from __future__ import unicode_literals

# local library
import inkex
from inkex.localization import inkex_gettext as _
from inkex.base import SvgOutputMixin

class hpglDecoder(SvgOutputMixin):
    def __init__(self, hpglString, options):
        """ options:
                "resolutionX":float
                "resolutionY":float
                "showMovements":bool
        """
        self.hpglString = hpglString
        self.options = options
        self.scaleX = options.resolutionX / 25.4  # dots/inch to dots/mm
        self.scaleY = options.resolutionY / 25.4  # dots/inch to dots/mm
        self.warning = ''
        self.textMovements = _("Movements")
        self.textPenNumber = _("Pen ")
        self.layers = {}
        self.oldCoordinates = (0.0, 297.0)

    def get_svg(self):
        """Generate an svg document from hgpl data"""
        actual_layer = 0
        # prepare document
        doc = self.get_template(width=210.0, height=297.0, unit='mm')
        svg = doc.getroot()
        svg.namedview.set('inkscape:document-units', 'mm')

        if self.options.showMovements:
            self.layers[0] = svg.add(inkex.Layer(self.textMovements))

        # cut stream into commands
        hpgl_data = self.hpglString.split(';')
        # if number of commands is under needed minimum, no data was found
        if len(hpgl_data) < 3:
            raise Exception('NO_HPGL_DATA')
        # decode commands into svg data
        for command in hpgl_data:
            if command.strip() != '':
                if command[:2] == 'IN' or command[:2] == 'FS' or command[:2] == 'VS':
                    # if Initialize, force or speed command ignore it
                    pass
                elif command[:2] == 'SP':
                    # if Select Pen command
                    actual_layer = int(command[2:])
                elif command[:2] == 'PU':
                    # if Pen Up command
                    self.parameters_to_path(svg, command[2:], 0, True)
                elif command[:2] == 'PD':
                    # if Pen Down command
                    self.parameters_to_path(svg, command[2:], actual_layer + 1, False)
                else:
                    self.warning = 'UNKNOWN_COMMANDS'
        return doc, self.warning

    def parameters_to_path(self, svg, parameters, layerNum, isPU):
        """split params and sanity check them"""
        parameters = parameters.strip().split(',')
        if parameters and len(parameters) % 2 == 0:
            for i, param in enumerate(parameters):
                # convert params to document units
                if i % 2 == 0:
                    parameters[i] = str(float(param) / self.scaleX)
                else:
                    parameters[i] = str(297.0 - (float(param) / self.scaleY))
            # create path and add it to the corresponding layer
            if not isPU or (self.options.showMovements and isPU):
                # create layer if it does not exist
                if layerNum not in self.layers:
                    label = self.textPenNumber + str(layerNum - 1)
                    self.layers[layerNum] = svg.add(inkex.Layer.new(label))

                path = 'M %f,%f L %s' % (self.oldCoordinates[0], self.oldCoordinates[1], ','.join(parameters))
                style = 'stroke:#' + ('ff0000' if isPU else '000000') + '; stroke-width:0.2; fill:none;'
                self.layers[layerNum].add(inkex.PathElement(d=path, style=style))
            self.oldCoordinates = (float(parameters[-2]), float(parameters[-1]))



# coding=utf-8
# COPYRIGHT
#
# pylint: disable=invalid-name
#
"""
Depreicated simplepath replacements with documentation
"""

import math
from inkex.deprecated import deprecate, DeprecatedDict
from inkex.transforms import Transform
from inkex.paths import Path

pathdefs = DeprecatedDict({
    'M':['L', 2, [float, float], ['x', 'y']],
    'L':['L', 2, [float, float], ['x', 'y']],
    'H':['H', 1, [float], ['x']],
    'V':['V', 1, [float], ['y']],
    'C':['C', 6, [float, float, float, float, float, float], ['x', 'y', 'x', 'y', 'x', 'y']],
    'S':['S', 4, [float, float, float, float], ['x', 'y', 'x', 'y']],
    'Q':['Q', 4, [float, float, float, float], ['x', 'y', 'x', 'y']],
    'T':['T', 2, [float, float], ['x', 'y']],
    'A':['A', 7, [float, float, float, int, int, float, float], ['r', 'r', 'a', 0, 's', 'x', 'y']],
    'Z':['L', 0, [], []]
})

@deprecate
def parsePath(d):
    """element.path.to_arrays()"""
    return Path(d).to_arrays()

@deprecate
def formatPath(a):
    """str(element.path) or str(Path(array))"""
    return str(Path(a))

@deprecate
def translatePath(p, x, y):
    """Path(array).translate(x, y)"""
    p[:] = Path(p).translate(x, y).to_arrays()

@deprecate
def scalePath(p, x, y):
    """Path(array).scale(x, y)"""
    p[:] = Path(p).scale(x, y).to_arrays()

@deprecate
def rotatePath(p, a, cx=0, cy=0):
    """Path(array).rotate(angle_degrees, (center_x, center_y))"""
    p[:] = Path(p).rotate(math.degrees(a), (cx, cy)).to_arrays()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import random
import inkex
from inkex import turtle as pturtle

class Lindenmayer(inkex.GenerateExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--order", type=int, default=3, help="number of iteration")
        pars.add_argument("--langle", type=float, default=16.0, help="angle for turning left")
        pars.add_argument("--rangle", type=float, default=16.0, help="angle for turning right")
        pars.add_argument("--step", type=float, default=25.0, help="step size")
        pars.add_argument("--randomizestep", type=float, default=0.0, help="randomize step")
        pars.add_argument("--randomizeangle", type=float, default=0.0, help="randomize angle")
        pars.add_argument("--axiom", default="++F", help="initial state of system")
        pars.add_argument("--rules", default="F=FF-[-F+F+F]+[+F-F-F]", help="replacement rules")
        self.stack = []
        self.turtle = pturtle.pTurtle()

    def iterate(self):
        self.rules = dict([map((lambda s: s.strip()), i.split("=")) for i in self.options.rules.upper().split(";") if i.count("=") == 1])
        string = self.__recurse(self.options.axiom.upper(), 0)
        self.__compose_path(string)
        return self.turtle.getPath()

    def __compose_path(self, string):
        self.turtle.pu()
        point = self.svg.namedview.center
        self.turtle.setpos(point)
        self.turtle.pd()
        for c in string:
            if c in 'ABCDEF':
                self.turtle.pd()
                self.turtle.fd(self.options.step * (random.normalvariate(1.0, 0.01 * self.options.randomizestep)))
            elif c in 'GHIJKL':
                self.turtle.pu()
                self.turtle.fd(self.options.step * (random.normalvariate(1.0, 0.01 * self.options.randomizestep)))
            elif c == '+':
                self.turtle.lt(self.options.langle * (random.normalvariate(1.0, 0.01 * self.options.randomizeangle)))
            elif c == '-':
                self.turtle.rt(self.options.rangle * (random.normalvariate(1.0, 0.01 * self.options.randomizeangle)))
            elif c == '|':
                self.turtle.lt(180)
            elif c == '[':
                self.stack.append([self.turtle.getpos(), self.turtle.getheading()])
            elif c == ']':
                self.turtle.pu()
                pos, heading = self.stack.pop()
                self.turtle.setpos(pos)
                self.turtle.setheading(heading)

    def __recurse(self, rule, level):
        level_string = ''
        for c in rule:
            if level < self.options.order:
                try:
                    level_string = level_string + self.__recurse(self.rules[c], level + 1)
                except KeyError:
                    level_string = level_string + c
            else:
                level_string = level_string + c
        return level_string

    def generate(self):
        self.options.step = self.svg.unittouu(str(self.options.step) + 'px')
        sty = {'stroke-linejoin': 'miter', 'stroke-width': str(self.svg.unittouu('1px')),
             'stroke-opacity': '1.0', 'fill-opacity': '1.0',
             'stroke': '#000000', 'stroke-linecap': 'butt',
             'fill': 'none'}
        return inkex.PathElement(style=str(inkex.Style(sty)), d=self.iterate())


if __name__ == '__main__':
    Lindenmayer().run()



# coding=utf-8
#
# Copyright (C) 2007 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
Python barcode renderer for Code93 barcodes. Designed for use with Inkscape.
"""

from .Base import Barcode

PALLET = list('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ-. $/+%')
PALLET.append('($)')
PALLET.append('(/)')
PALLET.append('(+)')
PALLET.append('(%)')
PALLET.append('MARKER')

MAP = dict((PALLET[i], i) for i in range(len(PALLET)))


def get_map(array):
    """Extended ENCODE maps for full ASCII Code93"""
    result = {}
    pos = 10
    for char in array:
        result[chr(char)] = PALLET[pos]
        pos += 1
    return result


# MapA is eclectic, but B, C, D are all ASCII ranges
MAP_A = get_map([27, 28, 29, 30, 31, 59, 60, 61, 62, 63, 91, 92, 93, 94, 95,
                 123, 124, 125, 126, 127, 0, 64, 96, 127, 127, 127])  # %
MAP_B = get_map(range(1, 26))  # $
MAP_C = get_map(range(33, 58))  # /
MAP_D = get_map(range(97, 122))  # +

ENCODE = [
    '100010100', '101001000', '101000100', '101000010', '100101000',
    '100100100', '100100010', '101010000', '100010010', '100001010',
    '110101000', '110100100', '110100010', '110010100', '110010010',
    '110001010', '101101000', '101100100', '101100010', '100110100',
    '100011010', '101011000', '101001100', '101000110', '100101100',
    '100010110', '110110100', '110110010', '110101100', '110100110',
    '110010110', '110011010', '101101100', '101100110', '100110110',
    '100111010', '100101110', '111010100', '111010010', '111001010',
    '101101110', '101110110', '110101110', '100100110', '111011010',
    '111010110', '100110010', '101011110', ''
]


class Code93(Barcode):
    def encode(self, text):
        # start marker
        bits = ENCODE[MAP.get('MARKER', -1)]

        # Extend to ASCII charset ( return Array )
        text = self.encode_ascii(text)

        # Calculate the checksums
        text.append(self.checksum(text, 20))  # C
        text.append(self.checksum(text, 15))  # K

        # Now convert text into the ENCODE bits (black and white stripes)
        for char in text:
            bits = bits + ENCODE[MAP.get(char, -1)]

        # end marker and termination bar
        return bits + ENCODE[MAP.get('MARKER', -1)] + '1'

    def checksum(self, text, mod):
        """Generate a code 93 checksum"""
        weight = len(text) % mod
        check = 0
        for char in text:
            check = check + (MAP[char] * weight)
            # Reset the weight is required
            weight -= 1
            if weight == 0:
                weight = mod

        return PALLET[check % 47]

    # Some characters need re-ENCODE into the code93 specification
    def encode_ascii(self, text):
        result = []
        for char in text:
            if char in MAP:
                result.append(char)
            elif char in MAP_A:
                result.append('(%)')
                result.append(MAP_A[char])
            elif char in MAP_B:
                result.append('($)')
                result.append(MAP_B[char])
            elif char in MAP_C:
                result.append('(/)')
                result.append(MAP_C[char])
            elif char in MAP_D:
                result.append('(+)')
                result.append(MAP_D[char])
        return result



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2007 John Beard john.j.beard@gmail.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This extension allows you to draw a Cartesian grid in Inkscape.

There is a wide range of options including subdivision, subsubdivions and
logarithmic scales. Custom line widths are also possible.

All elements are grouped with similar elements (eg all x-subdivs)
"""

from math import log

import inkex
from inkex import Group, PathElement, Rectangle

def draw_line(x1, y1, x2, y2, width, name, parent):
    """Draw an SVG line"""
    line = parent.add(PathElement())
    line.style = {'stroke': '#000000', 'stroke-width': str(width), 'fill': 'none'}
    line.path = 'M {},{} L {},{}'.format(x1, y1, x2, y2)
    line.label = name


def draw_rect(x, y, w, h, width, fill, name, parent):
    """Draw an SVG Rectangle"""
    rect = parent.add(Rectangle(x=str(x), y=str(y), width=str(w), height=str(h)))
    rect.style = {'stroke': '#000000', 'stroke-width': str(width), 'fill': fill}
    rect.label = name


class GridCartesian(inkex.GenerateExtension):
    def add_arguments(self, pars):
        pars.add_argument("--border_th", type=float, default=3)
        pars.add_argument("--border_th_unit", default="cm")
        pars.add_argument("--tab", default="x_tab")
        pars.add_argument("--x_divs", type=int, default=6)
        pars.add_argument("--dx", type=float, default=100.0)
        pars.add_argument("--dx_unit", default="cm")
        pars.add_argument("--x_subdivs", type=int, default=2)
        pars.add_argument("--x_log", type=inkex.Boolean, default="false")
        pars.add_argument("--x_subsubdivs", type=int, default=5)
        pars.add_argument("--x_half_freq", type=int, default=4)
        pars.add_argument("--x_divs_th", type=float, default=2)
        pars.add_argument("--x_subdivs_th", type=float, default=1)
        pars.add_argument("--x_subsubdivs_th", type=float, default=0.3)
        pars.add_argument("--x_div_unit", default="cm")
        pars.add_argument("--y_divs", type=int, default=5)
        pars.add_argument("--dy", type=float, default=100.0)
        pars.add_argument("--dy_unit", default="cm")
        pars.add_argument("--y_subdivs", type=int, default=1)
        pars.add_argument("--y_log", type=inkex.Boolean, default="false")
        pars.add_argument("--y_subsubdivs", type=int, default=5)
        pars.add_argument("--y_half_freq", type=int, default=4)
        pars.add_argument("--y_divs_th", type=float, default=2)
        pars.add_argument("--y_subdivs_th", type=float, default=1)
        pars.add_argument("--y_subsubdivs_th", type=float, default=0.3)
        pars.add_argument("--y_div_unit", default="cm")

    def generate(self):
        self.options.border_th = self.svg.unittouu(str(self.options.border_th) + self.options.border_th_unit)

        self.options.dx = self.svg.unittouu(str(self.options.dx) + self.options.dx_unit)
        self.options.x_divs_th = self.svg.unittouu(str(self.options.x_divs_th) + self.options.x_div_unit)
        self.options.x_subdivs_th = self.svg.unittouu(str(self.options.x_subdivs_th) + self.options.x_div_unit)
        self.options.x_subsubdivs_th = self.svg.unittouu(str(self.options.x_subsubdivs_th) + self.options.x_div_unit)

        self.options.dy = self.svg.unittouu(str(self.options.dy) + self.options.dy_unit)
        self.options.y_divs_th = self.svg.unittouu(str(self.options.y_divs_th) + self.options.y_div_unit)
        self.options.y_subdivs_th = self.svg.unittouu(str(self.options.y_subdivs_th) + self.options.y_div_unit)
        self.options.y_subsubdivs_th = self.svg.unittouu(str(self.options.y_subsubdivs_th) + self.options.y_div_unit)

        # find the pixel dimensions of the overall grid
        ymax = self.options.dy * self.options.y_divs
        xmax = self.options.dx * self.options.x_divs

        # Embed grid in group
        # Put in in the centre of the current view

        grid = Group.new("GridCartesian:X{0.x_divs}:Y{0.y_divs}".format(self.options))

        (pos_x, pos_y) = self.svg.namedview.center
        grid.transform.add_translate(pos_x - xmax / 2.0, pos_y - ymax / 2.0)

        # Group for major x gridlines
        majglx = grid.add(Group.new("MajorXGridlines"))
        # Group for major y gridlines
        majgly = grid.add(Group.new("MajorYGridlines"))

        # Group for minor x gridlines
        if self.options.x_subdivs > 1:  # if there are any minor x gridlines
            minglx = grid.add(Group.new("MinorXGridlines"))

        # Group for subminor x gridlines
        if self.options.x_subsubdivs > 1:  # if there are any minor minor x gridlines
            mminglx = grid.add(Group.new("SubMinorXGridlines"))

        # Group for minor y gridlines
        if self.options.y_subdivs > 1:  # if there are any minor y gridlines
            mingly = grid.add(Group.new("MinorYGridlines"))

        # Group for subminor y gridlines
        if self.options.y_subsubdivs > 1:  # if there are any minor minor x gridlines
            mmingly = grid.add(Group.new("SubMinorYGridlines"))

        draw_rect(0, 0, xmax, ymax, self.options.border_th,
                      'none', 'Border', grid)  # border rectangle

        # DO THE X DIVISIONS======================================
        sd = self.options.x_subdivs  # sub divs per div
        ssd = self.options.x_subsubdivs  # subsubdivs per subdiv

        for i in range(0, self.options.x_divs):  # Major x divisions
            if i > 0:  # don't draw first line (we made a proper border)
                draw_line(self.options.dx * i, 0,
                              self.options.dx * i, ymax,
                              self.options.x_divs_th,
                              'MajorXDiv' + str(i), majglx)

            if self.options.x_log:  # log x subdivs
                for j in range(1, sd):
                    if j > 1:  # the first loop is only for subsubdivs
                        draw_line(self.options.dx * (i + log(j, sd)), 0,
                                      self.options.dx * (i + log(j, sd)), ymax,
                                      self.options.x_subdivs_th,
                                      'MinorXDiv' + str(i) + ':' + str(j), minglx)

                    for k in range(1, ssd):  # subsub divs
                        if (j <= self.options.x_half_freq) or (k % 2 == 0):  # only draw half the subsubdivs past the half-freq point
                            if (ssd % 2 > 0) and (j > self.options.y_half_freq):  # half frequency won't work with odd numbers of subsubdivs,
                                ssd2 = ssd + 1  # make even
                            else:
                                ssd2 = ssd  # no change
                            draw_line(self.options.dx * (i + log(j + k / float(ssd2), sd)), 0,
                                          self.options.dx * (i + log(j + k / float(ssd2), sd)), ymax,
                                          self.options.x_subsubdivs_th, 'SubminorXDiv' + str(i) + ':' + str(j) + ':' + str(k), mminglx)

            else:  # linear x subdivs
                for j in range(0, sd):
                    if j > 0:  # not for the first loop (this loop is for the subsubdivs before the first subdiv)
                        draw_line(self.options.dx * (i + j / float(sd)), 0,
                                      self.options.dx * (i + j / float(sd)), ymax,
                                      self.options.x_subdivs_th,
                                      'MinorXDiv' + str(i) + ':' + str(j), minglx)

                    for k in range(1, ssd):  # subsub divs
                        draw_line(self.options.dx * (i + (j * ssd + k) / (float(sd) * ssd)), 0,
                                      self.options.dx * (i + (j * ssd + k) / (float(sd) * ssd)), ymax,
                                      self.options.x_subsubdivs_th,
                                      'SubminorXDiv' + str(i) + ':' + str(j) + ':' + str(k), mminglx)

        # DO THE Y DIVISIONS========================================
        sd = self.options.y_subdivs  # sub divs per div
        ssd = self.options.y_subsubdivs  # subsubdivs per subdiv

        for i in range(0, self.options.y_divs):  # Major y divisions
            if i > 0:  # don't draw first line (we will make a border)
                draw_line(0, self.options.dy * i,
                              xmax, self.options.dy * i,
                              self.options.y_divs_th,
                              'MajorYDiv' + str(i), majgly)

            if self.options.y_log:  # log y subdivs
                for j in range(1, sd):
                    if j > 1:  # the first loop is only for subsubdivs
                        draw_line(0, self.options.dy * (i + 1 - log(j, sd)),
                                      xmax, self.options.dy * (i + 1 - log(j, sd)),
                                      self.options.y_subdivs_th,
                                      'MinorXDiv' + str(i) + ':' + str(j), mingly)

                    for k in range(1, ssd):  # subsub divs
                        if (j <= self.options.y_half_freq) or (k % 2 == 0):  # only draw half the subsubdivs past the half-freq point
                            if (ssd % 2 > 0) and (j > self.options.y_half_freq):  # half frequency won't work with odd numbers of subsubdivs,
                                ssd2 = ssd + 1
                            else:
                                ssd2 = ssd  # no change
                            draw_line(0, self.options.dx * (i + 1 - log(j + k / float(ssd2), sd)),
                                          xmax, self.options.dx * (i + 1 - log(j + k / float(ssd2), sd)),
                                          self.options.y_subsubdivs_th,
                                          'SubminorXDiv' + str(i) + ':' + str(j) + ':' + str(k), mmingly)
            else:  # linear y subdivs
                for j in range(0, self.options.y_subdivs):
                    if j > 0:  # not for the first loop (this loop is for the subsubdivs before the first subdiv)
                        draw_line(0, self.options.dy * (i + j / float(sd)),
                                      xmax, self.options.dy * (i + j / float(sd)),
                                      self.options.y_subdivs_th,
                                      'MinorXYiv' + str(i) + ':' + str(j), mingly)

                    for k in range(1, ssd):  # subsub divs
                        draw_line(0, self.options.dy * (i + (j * ssd + k) / (float(sd) * ssd)),
                                      xmax, self.options.dy * (i + (j * ssd + k) / (float(sd) * ssd)),
                                      self.options.y_subsubdivs_th,
                                      'SubminorXDiv' + str(i) + ':' + str(j) + ':' + str(k), mmingly)

        return grid


if __name__ == '__main__':
    GridCartesian().run()



# coding=utf-8
#
# Copyright (C) 2020 Martin Owens <doctormo@geek-2.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Import this inkex module if you are using the extension manager
and need virtualenv dependencies. Things like gobject, numpy etc

Always import *before* anything else.
"""

import os
import sys

def get_bin(fname):
    """Get a virtualenv binary for execution, returns full filename"""
    for path in sys.path:
        for script in [fname, os.path.join('bin', fname)]:
            result = os.path.abspath(os.path.join(path, script))
            if os.path.isfile(result):
                return result
    return None

def activate_virtualenv():
    """ 
    The python that inkscape uses and the python installed into the virtualenv
    are different pythons with different libs. To give access to dependencies
    that are installed within the virtualenv, we activate the available venv.
    """
    activate_this = get_bin('activate_this.py')
    if activate_this and os.path.isfile(activate_this):
        with open(activate_this, 'r') as fhl:
            exec(fhl.read(), dict(__file__=activate_this))
        return

activate_virtualenv()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2005,2007 Aaron Spike, aaron@ekips.org
# - template dxf_outlines.dxf added Feb 2008 by Alvin Penner, penner@vaxxine.com
# - layers, transformation, flattening added April 2008 by Bob Cook, bob@bobcookdev.com
# - added support for dxf R12, Nov. 2008 by Switcher
# - brought together to replace ps2edit version 2018 by Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from __future__ import absolute_import, print_function, unicode_literals

import re
import inkex
from inkex.bezier import cspsubdiv

r12_header = ''' 0 
SECTION
 2 
HEADER
 9 
$ACADVER
 1 
AC1009
 9 
$EXTMIN
 10 
 0 
 20 
 0 
 9 
$EXTMAX
 10 
 8.5 
 20 
 11 
 0 
ENDSEC
 0 
SECTION
 2 
ENTITIES
'''

r12_footer = ''' 0 
ENDSEC
 0 
EOF'''


class DxfTwelve(inkex.OutputExtension):
    """Create dxf12 output from the svg"""
    def __init__(self):
        super(DxfTwelve, self).__init__()
        self.handle = 255
        self.flatness = 0.1

    def dxf_add(self, line):
        self._stream.write(line.encode('utf-8'))

    def dxf_insert_code(self, code, value):
        self.dxf_add(code + "\n" + value + "\n")

    def dxf_line(self, layer, csp):
        self.dxf_insert_code('0', 'LINE')
        self.dxf_insert_code('8', layer)
        # self.dxf_insert_code(  '62', '1' )  #Change the Line Color
        self.dxf_insert_code('10', '{:f}'.format(csp[0][0]))
        self.dxf_insert_code('20', '{:f}'.format(csp[0][1]))
        self.dxf_insert_code('11', '{:f}'.format(csp[1][0]))
        self.dxf_insert_code('21', '{:f}'.format(csp[1][1]))

    def dxf_path_to_lines(self, layer, p):
        f = self.flatness
        is_flat = 0
        while is_flat < 1:
            try:
                cspsubdiv(p, self.flatness)
                is_flat = 1
            except:
                f += 0.1

        for sub in p:
            for i in range(len(sub) - 1):
                self.handle += 1
                s = sub[i]
                e = sub[i + 1]
                self.dxf_line(layer, [s[1], e[1]])

    def dxf_path_to_point(self, layer, p):
        bbox = inkex.Path(p).bounding_box() or inkex.BoundingBox(0, 0)
        x, y = bbox.center
        self.dxf_point(layer, x, y)

    def save(self, stream):
        self._stream = stream
        self.dxf_insert_code('999', '"DXF R12 Output" (www.mydxf.blogspot.com)')
        self.dxf_add(r12_header)

        scale = 25.4 / 90.0
        h = self.svg.height

        path = '//svg:path'
        for node in self.svg.xpath(path):

            layer = node.getparent().label
            if layer is None:
                layer = 'Layer 1'

            node.transform *= inkex.Transform([[scale, 0, 0], [0, -scale, h * scale]])
            node.apply_transform()
            path = node.path.to_superpath()

            if re.search('drill$', layer, re.I) is None:
                # if layer == 'Brackets Drill':
                self.dxf_path_to_lines(layer, path)
            else:
                self.dxf_path_to_point(layer, path)

        self.dxf_add(r12_footer)


if __name__ == '__main__':
    DxfTwelve().run()



# coding=utf-8
#
# Copyright (C) 2018-2019 Martin Owens
#               2019 Thomas Holder
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
All Inkscape extensions should come with tests. This package provides you with
the tools needed to create tests and thus ensure that your extension continues
to work with future versions of Inkscape, the "inkex" python modules, and other
python and non-python tools you may use.

Make sure your extension is a python extension and is using the `inkex.generic`
base classes. These provide the greatest amount of functionality for testing.

You should start by creating a folder in your repository called `tests` with
an empty file inside called `__init__.py` to turn it into a module folder.

For each of your extensions, you should create a file called
`test_{extension_name}.py` where the name reflects the name of your extension.

There are two types of tests:

    1. Full-process Comparison tests - These are tests which invoke your
           extension with various arguments and attempt to compare the
           output to a known good reference. These are useful for testing
           that your extension would work if it was used in Inkscape.

           Good example of writing comparison tests can be found in the
           Inkscape core repository, each test which inherits from
           the ComparisonMixin class is running comparison tests.

    2. Unit tests - These are individual test functions which call out to
           specific functions within your extension. These are typical
           python unit tests and many good python documents exist
           to describe how to write them well. For examples here you
           can find the tests that test the inkex modules themselves
           to be the most instructive.

When running a test, it will cause a certain fraction of the code within the
extension to execute. This fraction called it's **coverage** and a higher
coverage score indicates that your test is better at exercising the various
options, features, and branches within your code.

Generating comparison output can be done using the EXPORT_COMPARE environment
variable when calling pytest. For example:

    EXPORT_COMPARE=1 pytest tests/test_my_specific_test.py

This will create files in `tests/data/refs/*.out.export` and these files should
be manually checked to make sure they are correct before being renamed and stripped
of the `.export` suffix. pytest should then be re-run to confirm before
committing to the repository.
"""

import os
import re
import sys
import shutil
import tempfile
import hashlib
import random
import uuid
from typing import List, Union, Tuple

from io import BytesIO, StringIO
import xml.etree.ElementTree as xml

from unittest import TestCase as BaseCase
from inkex.base import InkscapeExtension

from .. import Transform
from ..utils import to_bytes
from .xmldiff import xmldiff
from .mock import MockCommandMixin, Capture

if False: # pylint: disable=using-constant-test
    from typing import Type, List
    from .filters import Compare


class NoExtension(InkscapeExtension):  # pylint: disable=too-few-public-methods
    """Test case must specify 'self.effect_class' to assertEffect."""

    def __init__(self, *args, **kwargs): # pylint: disable=super-init-not-called
        raise NotImplementedError(self.__doc__)

    def run(self, args=None, output=None):
        """Fake run"""
        pass


class TestCase(MockCommandMixin, BaseCase):
    """
    Base class for all effects tests, provides access to data_files and test_without_parameters
    """
    effect_class = NoExtension # type: Type[InkscapeExtension]
    effect_name = property(lambda self: self.effect_class.__module__)

    # If set to true, the output is not expected to be the stdout SVG document, but rather
    # text or a message sent to the stderr, this is highly weird. But sometimes happens.
    stderr_output = False
    stdout_protect = True
    stderr_protect = True

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._temp_dir = None
        self._effect = None

    def setUp(self): # pylint: disable=invalid-name
        """Make sure every test is seeded the same way"""
        self._effect = None
        super().setUp()
        random.seed(0x35f)

    def tearDown(self):
        super().tearDown()
        if self._temp_dir and os.path.isdir(self._temp_dir):
            shutil.rmtree(self._temp_dir)

    @classmethod
    def __file__(cls):
        """Create a __file__ property which acts much like the module version"""
        return os.path.abspath(sys.modules[cls.__module__].__file__)

    @classmethod
    def _testdir(cls):
        """Get's the folder where the test exists (so data can be found)"""
        return os.path.dirname(cls.__file__())

    @classmethod
    def rootdir(cls):
        """Return the full path to the extensions directory"""
        return os.path.dirname(cls._testdir())

    @classmethod
    def datadir(cls):
        """Get the data directory (can be over-ridden if needed)"""
        return os.path.join(cls._testdir(), 'data')

    @property
    def tempdir(self):
        """Generate a temporary location to store files"""
        if self._temp_dir is None:
            self._temp_dir = tempfile.mkdtemp(prefix='inkex-tests-')
        if not os.path.isdir(self._temp_dir):
            raise IOError("The temporary directory has disappeared!")
        return self._temp_dir

    def temp_file(self, prefix='file-', template='{prefix}{name}{suffix}', suffix='.tmp'):
        """Generate the filename of a temporary file"""
        filename = template.format(prefix=prefix, suffix=suffix, name=uuid.uuid4().hex)
        return os.path.join(self.tempdir, filename)

    @classmethod
    def data_file(cls, filename, *parts):
        """Provide a data file from a filename, can accept directories as arguments."""
        if os.path.isabs(filename):
            # Absolute root was passed in, so we trust that (it might be a tempdir)
            full_path = os.path.join(filename, *parts)
        else:
            # Otherwise we assume it's relative to the test data dir.
            full_path = os.path.join(cls.datadir(), filename, *parts)

        if not os.path.isfile(full_path):
            raise IOError(f"Can't find test data file: {full_path}")
        return full_path

    @property
    def empty_svg(self):
        """Returns a common minimal svg file"""
        return self.data_file('svg', 'default-inkscape-SVG.svg')

    def assertAlmostTuple(self, found, expected, precision=8): # pylint: disable=invalid-name
        """
        Floating point results may vary with computer architecture; use
        assertAlmostEqual to allow a tolerance in the result.
        """
        self.assertEqual(len(found), len(expected))
        for fon, exp in zip(found, expected):
            self.assertAlmostEqual(fon, exp, precision)

    def assertEffectEmpty(self, effect, **kwargs):  # pylint: disable=invalid-name
        """Assert calling effect without any arguments"""
        self.assertEffect(effect=effect, **kwargs)

    def assertEffect(self, *filename, **kwargs):  # pylint: disable=invalid-name
        """Assert an effect, capturing the output to stdout.

           filename should point to a starting svg document, default is empty_svg
        """
        data_file = self.data_file(*filename) if filename else self.empty_svg

        os.environ['DOCUMENT_PATH'] = data_file
        args = [data_file] + list(kwargs.pop('args', []))
        args += ['--{}={}'.format(*kw) for kw in kwargs.items()]

        effect = kwargs.pop('effect', self.effect_class)()

        # Output is redirected to this string io buffer
        if self.stderr_output:
            with Capture('stderr') as stderr:
                effect.run(args, output=BytesIO())
                effect.test_output = stderr
        else:
            output = BytesIO()
            with Capture('stdout', kwargs.get('stdout_protect', self.stdout_protect)) as stdout:
                with Capture('stderr', kwargs.get('stderr_protect', self.stderr_protect)) as stderr:
                    effect.run(args, output=output)
                    self.assertEqual('', stdout.getvalue(), "Extra print statements detected")
                    self.assertEqual('', stderr.getvalue(), "Extra error or warnings detected")
            effect.test_output = output

        if os.environ.get('FAIL_ON_DEPRECATION', False):
            warnings = getattr(effect, 'warned_about', set())
            effect.warned_about = set()  # reset for next test
            self.assertFalse(warnings, "Deprecated API is still being used!")

        return effect

    def assertDeepAlmostEqual(self, first, second, places=None, msg=None, delta=None):
        if delta is None and places is None:
            places = 7
        if isinstance(first, (list, tuple)):
            assert len(first) == len(second)
            for (f, s) in zip(first, second):
                self.assertDeepAlmostEqual(f, s, places, msg, delta)
        else:
            self.assertAlmostEqual(first, second, places, msg, delta)

    def assertTransformEqual(self, lhs, rhs, places=7):
        """Assert that two transform expressions evaluate to the same
        transformation matrix.
        """
        self.assertAlmostTuple(tuple(Transform(lhs).to_hexad()),
                               tuple(Transform(rhs).to_hexad()), places)

    @property
    def effect(self):
        """Generate an effect object"""
        if self._effect is None:
            self._effect = self.effect_class()
        return self._effect

class InkscapeExtensionTestMixin:
    """Automatically setup self.effect for each test and test with an empty svg"""
    def setUp(self): # pylint: disable=invalid-name
        """Check if there is an effect_class set and create self.effect if it is"""
        super(InkscapeExtensionTestMixin, self).setUp()
        if self.effect_class is None:
            self.skipTest('self.effect_class is not defined for this this test')

    def test_default_settings(self):
        """Extension works with empty svg file"""
        self.effect.run([self.empty_svg])

class ComparisonMixin:
    """
    Add comparison tests to any existing test suite.
    """
    # This input svg file sent to the extension (if any)
    compare_file: Union[List[str], Tuple[str], str] = 'svg/shapes.svg'
    # The ways in which the output is filtered for comparision (see filters.py)
    compare_filters = [] # type: List[Compare]
    # If true, the filtered output will be saved and only applied to the
    # extension output (and not to the reference file)
    compare_filter_save = False
    # A list of comparison runs, each entry will cause the extension to be run.
    comparisons = [
        (),
        ('--id=p1', '--id=r3'),
    ]

    def test_all_comparisons(self):
        """Testing all comparisons"""
        if not isinstance(self.compare_file, (list, tuple)):
            self._test_comparisons(self.compare_file)
        else:
            for compare_file in self.compare_file:
                self._test_comparisons(
                    compare_file,
                    addout=os.path.basename(compare_file)
                )

    def _test_comparisons(self, compare_file, addout=None):
        for args in self.comparisons:
            self.assertCompare(
                compare_file,
                self.get_compare_outfile(args, addout),
                args,
            )

    def assertCompare(self, infile, outfile, args): #pylint: disable=invalid-name
        """
        Compare the output of a previous run against this one.

         - infile: The filename of the pre-processed svg (or other type of file)
         - outfile: The filename of the data we expect to get, if not set
                    the filename will be generated from the effect name and kwargs.
         - args: All the arguments to be passed to the effect run

        """
        effect = self.assertEffect(infile, args=args)

        if outfile is None:
            outfile = self.get_compare_outfile(args)

        if not os.path.isfile(outfile):
            raise IOError(f"Comparison file {outfile} not found")

        data_a = effect.test_output.getvalue()
        if os.environ.get('EXPORT_COMPARE', False):
            with open(outfile + '.export', 'wb') as fhl:
                if sys.version_info[0] == 3 and isinstance(data_a, str):
                    data_a = data_a.encode('utf-8')
                fhl.write(self._apply_compare_filters(data_a, True))
                print(f"Written output: {outfile}.export")

        data_a = self._apply_compare_filters(data_a)

        with open(outfile, 'rb') as fhl:
            data_b = self._apply_compare_filters(fhl.read(), False)

        if isinstance(data_a, bytes) and isinstance(data_b, bytes) \
            and data_a.startswith(b'<') and data_b.startswith(b'<'):
            # Late importing
            diff_xml, delta = xmldiff(data_a, data_b)
            if not delta and not os.environ.get('EXPORT_COMPARE', False):
                print('The XML is different, you can save the output using the EXPORT_COMPARE=1'\
                      ' envionment variable. This will save the compared file as a ".output" file'\
                      ' next to the reference file used in the test.\n')
            diff = f"SVG Differences: {outfile}\n\n"
            if os.environ.get('XML_DIFF', False):
                diff = '<- ' + diff_xml
            else:
                for x, (value_a, value_b) in enumerate(delta):
                    try:
                        # Take advantage of better text diff in testcase's own asserts.
                        self.assertEqual(value_a, value_b)
                    except AssertionError as err:
                        diff += f" {x}. {str(err)}\n"
            self.assertTrue(delta, diff)
        else:
            # compare any content (non svg)
            self.assertEqual(data_a, data_b)

    def _apply_compare_filters(self, data, is_saving=None):
        data = to_bytes(data)
        # Applying filters flips depending if we are saving the filtered content
        # to disk, or filtering during the test run. This is because some filters
        # are destructive others are useful for diagnostics.
        if is_saving is self.compare_filter_save or is_saving is None:
            for cfilter in self.compare_filters:
                data = cfilter(data)
        return data

    def get_compare_outfile(self, args, addout=None):
        """Generate an output file for the arguments given"""
        if addout is not None:
            args = list(args) + [str(addout)]
        opstr = '__'.join(args)\
                    .replace(self.tempdir, 'TMP_DIR')\
                    .replace(self.datadir(), 'DAT_DIR')
        opstr = re.sub(r'[^\w-]', '__', opstr)
        if opstr:
            if len(opstr) > 127:
                # avoid filename-too-long error
                opstr = hashlib.md5(opstr.encode('latin1')).hexdigest()
            opstr = '__' + opstr
        return self.data_file("refs", f"{self.effect_name}{opstr}.out")



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2008-2009 Alvin Penner, penner@vaxxine.com
#               2009, Christian Mayer, inkscape@christianmayer.de
#               2020, MartinOwens, doctormo@geek-2.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Input a DXF file >= (AutoCAD Release 13 == AC1012)
"""

import os
import re
import sys
import math

from collections import defaultdict
from urllib.parse import quote
from lxml import etree

import inkex

COLORS = [
    'PAD',
    '#FF0000', '#FFFF00', '#00FF00', '#00FFFF', '#0000FF', '#FF00FF', '#000000', '#808080',
    '#C0C0C0', '#FF0000', '#FF7F7F', '#CC0000', '#CC6666', '#990000', '#994C4C', '#7F0000',
    '#7F3F3F', '#4C0000', '#4C2626', '#FF3F00', '#FF9F7F', '#CC3300', '#CC7F66', '#992600',
    '#995F4C', '#7F1F00', '#7F4F3F', '#4C1300', '#4C2F26', '#FF7F00', '#FFBF7F', '#CC6600',
    '#CC9966', '#994C00', '#99724C', '#7F3F00', '#7F5F3F', '#4C2600', '#4C3926', '#FFBF00',
    '#FFDF7F', '#CC9900', '#CCB266', '#997200', '#99854C', '#7F5F00', '#7F6F3F', '#4C3900',
    '#4C4226', '#FFFF00', '#FFFF7F', '#CCCC00', '#CCCC66', '#989800', '#98984C', '#7F7F00',
    '#7F7F3F', '#4C4C00', '#4C4C26', '#BFFF00', '#DFFF7F', '#99CC00', '#B2CC66', '#729800',
    '#85984C', '#5F7F00', '#6F7F3F', '#394C00', '#424C26', '#7FFF00', '#BFFF7F', '#66CC00',
    '#99CC66', '#4C9800', '#72984C', '#3F7F00', '#5F7F3F', '#264C00', '#394C26', '#3FFF00',
    '#9FFF7F', '#33CC00', '#7FCC66', '#269800', '#5F984C', '#1F7F00', '#4F7F3F', '#134C00',
    '#2F4C26', '#00FF00', '#7FFF7F', '#00CC00', '#66CC66', '#009800', '#4C984C', '#007F00',
    '#3F7F3F', '#004C00', '#264C26', '#00FF3F', '#7FFF9F', '#00CC33', '#66CC7F', '#009826',
    '#4C985F', '#007F1F', '#3F7F4F', '#004C13', '#264C2F', '#00FF7F', '#7FFFBF', '#00CC66',
    '#66CC99', '#00984C', '#4C9872', '#007F3F', '#3F7F5F', '#004C26', '#264C39', '#00FFBF',
    '#7FFFDF', '#00CC99', '#66CCB2', '#009872', '#4C9885', '#007F5F', '#3F7F6F', '#004C39',
    '#264C42', '#00FFFF', '#7FFFFF', '#00CCCC', '#66CCCC', '#009898', '#4C9898', '#007F7F',
    '#FF0000', '#FFFF00', '#00FF00', '#00FFFF', '#0000FF', '#FF00FF', '#000000', '#808080',
    '#C0C0C0', '#FF0000', '#FF7F7F', '#CC0000', '#CC6666', '#990000', '#994C4C', '#7F0000',
    '#7F3F3F', '#4C0000', '#4C2626', '#FF3F00', '#FF9F7F', '#CC3300', '#CC7F66', '#992600',
    '#995F4C', '#7F1F00', '#7F4F3F', '#4C1300', '#4C2F26', '#FF7F00', '#FFBF7F', '#CC6600',
    '#CC9966', '#994C00', '#99724C', '#7F3F00', '#7F5F3F', '#4C2600', '#4C3926', '#FFBF00',
    '#FFDF7F', '#CC9900', '#CCB266', '#997200', '#99854C', '#7F5F00', '#7F6F3F', '#4C3900',
    '#4C4226', '#FFFF00', '#FFFF7F', '#CCCC00', '#CCCC66', '#989800', '#98984C', '#7F7F00',
    '#7F7F3F', '#4C4C00', '#4C4C26', '#BFFF00', '#DFFF7F', '#99CC00', '#B2CC66', '#729800',
    '#85984C', '#5F7F00', '#6F7F3F', '#394C00', '#424C26', '#7FFF00', '#BFFF7F', '#66CC00',
    '#99CC66', '#4C9800', '#72984C', '#3F7F00', '#5F7F3F', '#264C00', '#394C26', '#3FFF00',
    '#9FFF7F', '#33CC00', '#7FCC66', '#269800', '#5F984C', '#1F7F00', '#4F7F3F', '#134C00',
    '#2F4C26', '#00FF00', '#7FFF7F', '#00CC00', '#66CC66', '#009800', '#4C984C', '#007F00',
    '#3F7F3F', '#004C00', '#264C26', '#00FF3F', '#7FFF9F', '#00CC33', '#66CC7F', '#009826',
    '#4C985F', '#007F1F', '#3F7F4F', '#004C13', '#264C2F', '#00FF7F', '#7FFFBF', '#00CC66',
    '#66CC99', '#00984C', '#4C9872', '#007F3F', '#3F7F5F', '#004C26', '#264C39', '#00FFBF',
    '#7FFFDF', '#00CC99', '#66CCB2', '#009872', '#4C9885', '#007F5F', '#3F7F6F', '#004C39',
    '#264C42', '#00FFFF', '#7FFFFF', '#00CCCC', '#66CCCC', '#009898', '#4C9898', '#007F7F',
    '#3F7F7F', '#004C4C', '#264C4C', '#00BFFF', '#7FDFFF', '#0099CC', '#66B2CC', '#007298',
    '#4C8598', '#005F7F', '#3F6F7F', '#00394C', '#26424C', '#007FFF', '#7FBFFF', '#0066CC',
    '#6699CC', '#004C98', '#4C7298', '#003F7F', '#3F5F7F', '#00264C', '#26394C', '#003FFF',
    '#7F9FFF', '#0033CC', '#667FCC', '#002698', '#4C5F98', '#001F7F', '#3F4F7F', '#00134C',
    '#262F4C', '#0000FF', '#7F7FFF', '#0000CC', '#6666CC', '#000098', '#4C4C98', '#00007F',
    '#3F3F7F', '#00004C', '#26264C', '#3F00FF', '#9F7FFF', '#3300CC', '#7F66CC', '#260098',
    '#5F4C98', '#1F007F', '#4F3F7F', '#13004C', '#2F264C', '#7F00FF', '#BF7FFF', '#6600CC',
    '#9966CC', '#4C0098', '#724C98', '#3F007F', '#5F3F7F', '#26004C', '#39264C', '#BF00FF',
    '#DF7FFF', '#9900CC', '#B266CC', '#720098', '#854C98', '#5F007F', '#6F3F7F', '#39004C',
    '#42264C', '#FF00FF', '#FF7FFF', '#CC00CC', '#CC66CC', '#980098', '#984C98', '#7F007F',
    '#7F3F7F', '#4C004C', '#4C264C', '#FF00BF', '#FF7FDF', '#CC0099', '#CC66B2', '#980072',
    '#984C85', '#7F005F', '#7F3F6F', '#4C0039', '#4C2642', '#FF007F', '#FF7FBF', '#CC0066',
    '#CC6699', '#98004C', '#984C72', '#7F003F', '#7F3F5F', '#4C0026', '#4C2639', '#FF003F',
    '#FF7F9F', '#CC0033', '#CC667F', '#980026', '#984C5F', '#7F001F', '#7F3F4F', '#4C0013',
    '#4C262F', '#333333', '#5B5B5B', '#848484', '#ADADAD', '#D6D6D6', '#FFFFFF'
]

def get_rgbcolor(dxfcolor):
    if dxfcolor in range(1,len(COLORS)):
        rgbcolor = COLORS[dxfcolor]
    else:
        rgbcolor = '#000000'
    return rgbcolor

class ValueConstruct(defaultdict):
    """Store values from the DXF and provide them as named attributes"""
    values = {
        '1': ('text', 'default'),
        '2': ('tag', 'block_name'),
        '3': ('mtext',),
        '6': ('line_type',),
        '8': ('layer_name',),
        '10': ('x1',),
        '11': ('x2',),
        '13': ('x3',),
        '14': ('x4',),
        '20': ('y1',),
        '21': ('y2',),
        '23': ('y3',),
        '24': ('y4',),
        '40': ('scale', 'knots', 'radius', 'width_ratio'),
        '41': ('ellipse_a1', 'insert_scale_x'),
        '42': ('ellipse_a2', 'bulge', 'insert_scale_y'),
        '50': ('angle',),
        '51': ('angle2',),
        '62': ('color',),
        '70': ('fill', 'flags'),
        '72': ('edge_type',),
        '73': ('sweep',), # ccw
        '92': ('path_type',),
        '93': ('num_edges',),
        '230': ('extrude',),
        '370': ('line_weight',),
    }
    attrs = dict([(name, a) for a, b in values.items() for name in b])

    def __init__(self):
        super().__init__(list)

    @classmethod
    def is_valid(cls, key):
        return key in cls.values

    def __getattr__(self, attr):
        is_list = attr.endswith('_list')
        key = attr[:-5] if is_list else attr
        if key in self.attrs:
            ret = self[self.attrs[key]]
            if not attr.endswith('_list'):
                return ret[0]
            return ret
        if attr.startswith('has_'):
            key = attr[4:]
            if key in self.attrs:
                return self.attrs[key] in self
        raise AttributeError(f"Can't find dxf attribute '{key}' {attr}")

    def __setattr__(self, attr, value):
        if not attr in self.attrs:
            raise AttributeError(f"Can't set bad dxf attribute '{key}'")
        if not isinstance(value, list):
            value = [value]
        self[self.attrs[attr]] = value

    def adjust_coords(self, xmin, ymin, scale, extrude, height):
        """Adjust the x,y coordinates to fit on the page"""
        for xgrp in set(['10', '11', '13', '14']) & set(self):  # scale/reflect x values
            for i in range(len(self[xgrp])):
                self[xgrp][i] = scale * (extrude * self[xgrp][i] - xmin)
        for ygrp in set(['20', '21', '23', '24']) & set(self):  # scale y values
            for i in range(len(self[ygrp])):
                self[ygrp][i] = height - scale * (self[ygrp][i] - ymin)

export_viewport = False
export_endsec = False

def re_hex2unichar(m):
    return chr(int(m.group(1), 16))


def formatStyle(style):
    return str(inkex.Style(style))

def export_text(*args, **kwargs):
    return export_mtext(*args, **kwargs)

def export_mtext(vals):
    # mandatory group codes : (1 or 3, 10, 20) (text, x, y)
    if (vals.has_text or vals.has_mtext) and vals.has_x1 and vals.has_y1:
        x = vals.x1
        y = vals.y1
        # optional group codes : (21, 40, 50) (direction, text height mm, text angle)
        size = 12  # default fontsize in px
        if vals.has_scale:
            size = scale * textscale * vals.scale
        attribs = {'x': '%f' % x, 'y': '%f' % y, 'style': 'font-size: %.3fpx; fill: %s; font-family: %s' % (size, color, options.font)}
        angle = 0  # default angle in degrees
        if vals.has_angle:
            angle = vals.angle
            attribs.update({'transform': 'rotate (%f %f %f)' % (-angle, x, y)})
        elif vals.has_y2:
            if vals.y2 == 1.0:
                attribs.update({'transform': 'rotate (%f %f %f)' % (-90, x, y)})
            elif vals.y2 == -1.0:
                attribs.update({'transform': 'rotate (%f %f %f)' % (90, x, y)})
        node = layer.add(inkex.TextElement(**attribs))
        node.set('sodipodi:linespacing', '125%')
        text = ''
        if vals.has_mtext:
            text = ''.join(vals.mtext_list)
        if vals.has_text:
            text = vals.text
        found = text.find(r'\P')  # new line
        while found > -1:
            tspan = node.add(inkex.Tspan())
            tspan.set('sodipodi:role', 'line')
            tspan.text = text[:found]
            text = text[(found + 2):]
            found = text.find(r'\P')
        tspan = node.add(inkex.Tspan())
        tspan.set('sodipodi:role', 'line')
        tspan.text = text


def export_point(vals, w):
    # mandatory group codes : (10, 20) (x, y)
    if vals.has_x1 and vals.has_y1:
        if options.gcodetoolspoints:
            generate_gcodetools_point(vals.x1, vals.y1)
        else:
            generate_ellipse(vals.x1, vals.y1, w / 2, 0.0, 1.0, 0.0, 0.0)


def export_line(vals):
    """Draw a strait line from the dxf"""
    # mandatory group codes : (10, 11, 20, 21) (x1, x2, y1, y2)
    if vals.has_x1 and vals.has_x2 and vals.has_y1 and vals.has_y2:
        path = inkex.PathElement()
        path.style = style
        path.path = 'M %f,%f %f,%f' % (vals.x1, vals.y1, vals.x2, vals.y2)
        layer.add(path)


def export_spline(vals):
    # see : http://www.mactech.com/articles/develop/issue_25/schneider.html
    # mandatory group codes : (10, 20, 40, 70) (x[], y[], knots[], flags)
    if vals.has_flags and vals.has_knots and vals.x1_list \
            and len(vals.x1_list) == len(vals.y1_list):
        knots = vals.knots_list
        ctrls = len(vals.x1_list)
        if ctrls > 3 and len(knots) == ctrls + 4:  # cubic
            if ctrls > 4:
                for i in range(len(knots) - 5, 3, -1):
                    if knots[i] != knots[i - 1] and knots[i] != knots[i + 1]:
                        a0 = (knots[i] - knots[i - 2]) / (knots[i + 1] - knots[i - 2])
                        a1 = (knots[i] - knots[i - 1]) / (knots[i + 2] - knots[i - 1])
                        vals.x1_list.insert(i - 1, (1.0 - a1) * vals.x1_list[i - 2] + a1 * vals.x1_list[i - 1])
                        vals.y1_list.insert(i - 1, (1.0 - a1) * vals.y1_list[i - 2] + a1 * vals.y1_list[i - 1])
                        vals.x1_list[i - 2] = (1.0 - a0) * vals.x1_list[i - 3] + a0 * vals.x1_list[i - 2]
                        vals.y1_list[i - 2] = (1.0 - a0) * vals.y1_list[i - 3] + a0 * vals.y1_list[i - 2]
                        knots.insert(i, knots[i])
                for i in range(len(knots) - 6, 3, -2):
                    if knots[i] != knots[i + 2] and knots[i - 1] != knots[i + 1] and knots[i - 2] != knots[i]:
                        a1 = (knots[i] - knots[i - 1]) / (knots[i + 2] - knots[i - 1])
                        vals.x1_list.insert(i - 1, (1.0 - a1) * vals.x1_list[i - 2] + a1 * vals.x1_list[i - 1])
                        vals.y1_list.insert(i - 1, (1.0 - a1) * vals.y1_list[i - 2] + a1 * vals.y1_list[i - 1])
            ctrls = len(vals.x1_list)
            path = 'M %f,%f' % (vals.x1, vals.y1)
            for i in range(0, (ctrls - 1) // 3):
                path += ' C %f,%f %f,%f %f,%f' % (vals.x1_list[3 * i + 1], vals.y1_list[3 * i + 1], vals.x1_list[3 * i + 2], vals.y1_list[3 * i + 2], vals.x1_list[3 * i + 3], vals.y1_list[3 * i + 3])
            if vals.flags & 1:  # closed path
                path += ' z'
            attribs = {'d': path, 'style': style}
            etree.SubElement(layer, 'path', attribs)
        if ctrls == 3 and len(knots) == 6:  # quadratic
            path = 'M %f,%f Q %f,%f %f,%f' % (vals.x1, vals.y1, vals.x1_list[1], vals.y1_list[1], vals.x1_list[2], vals.y1_list[2])
            attribs = {'d': path, 'style': style}
            etree.SubElement(layer, 'path', attribs)
        if ctrls == 5 and len(knots) == 8:  # spliced quadratic
            path = 'M %f,%f Q %f,%f %f,%f Q %f,%f %f,%f' % (vals.x1, vals.y1, vals.x1_list[1], vals.y1_list[1], vals.x1_list[2], vals.y1_list[2], vals.x1_list[3], vals.y1_list[3], vals.x1_list[4], vals.y1_list[4])
            attribs = {'d': path, 'style': style}
            etree.SubElement(layer, 'path', attribs)


def export_circle(vals):
    # mandatory group codes : (10, 20, 40) (x, y, radius)
    if vals.has_x1 and vals.has_y1 and vals.has_radius:
        generate_ellipse(vals.x1, vals.y1, scale * vals.radius, 0.0, 1.0, 0.0, 0.0)

def export_arc(vals):
    # mandatory group codes : (10, 20, 40, 50, 51) (x, y, radius, angle1, angle2)
    if vals.has_x1 and vals.has_y1 and vals.has_radius and vals.has_angle and vals.has_angle2:
        generate_ellipse(vals.x1, vals.y1,
            scale * vals.radius, 0.0, 1.0, vals.angle * math.pi / 180.0,
            vals.angle2 * math.pi / 180.0)


def export_ellipse(vals):
    # mandatory group codes : (10, 11, 20, 21, 40, 41, 42) (xc, xm, yc, ym, width ratio, angle1, angle2)
    if vals.has_x1 and vals.has_x2 and vals.has_y1 and vals.has_y2 and \
            vals.has_width_ratio and vals.has_ellipse_a1 and vals.has_ellipse_a2:
        generate_ellipse(vals.x1, vals.y1, vals.x2, vals.y2, vals.width_ratio, vals.ellipse_a1, vals.ellipse_a2)


def export_leader(vals):
    # mandatory group codes : (10, 20) (x, y)
    if vals.has_x1 and vals.has_y1:
        if len(vals.x1_list) > 1 and len(vals.y1_list) == len(vals.x1_list):
            path = 'M %f,%f' % (vals.x1, vals.y1)
            for i in range(1, len(vals.x1_list)):
                path += ' %f,%f' % (vals.x1_list[i], vals.y1_list[i])
            attribs = {'d': path, 'style': style}
            etree.SubElement(layer, 'path', attribs)

def export_polyline(vals):
    return export_lwpolyline(vals)    

def export_lwpolyline(vals):
    # mandatory group codes : (10, 20, 70) (x, y, flags)
    if vals.has_x1 and vals.has_y1 and vals.has_flags:
        if len(vals.x1_list) > 1 and len(vals.y1_list) == len(vals.x1_list):
            # optional group codes : (42) (bulge)
            iseqs = 0
            ibulge = 0
            if vals.flags & 1:  # closed path
                seqs.append('20')
                vals.x1_list.append(vals.x1)
                vals.y1_list.append(vals.y1)
            while seqs[iseqs] != '20':
                iseqs += 1
            path = 'M %f,%f' % (vals.x1, vals.y1)
            xold = vals.x1
            yold = vals.y1
            for i in range(1, len(vals.x1_list)):
                bulge = 0
                iseqs += 1
                while seqs[iseqs] != '20':
                    if seqs[iseqs] == '42':
                        bulge = vals.bulge_list[ibulge]
                        ibulge += 1
                    iseqs += 1
                if bulge:
                    sweep = 0  # sweep CCW
                    if bulge < 0:
                        sweep = 1  # sweep CW
                        bulge = -bulge
                    large = 0  # large-arc-flag
                    if bulge > 1:
                        large = 1
                    r = math.sqrt((vals.x1_list[i] - xold) ** 2 + (vals.y1_list[i] - yold) ** 2)
                    r = 0.25 * r * (bulge + 1.0 / bulge)
                    path += ' A %f,%f 0.0 %d %d %f,%f' % (r, r, large, sweep, vals.x1_list[i], vals.y1_list[i])
                else:
                    path += ' L %f,%f' % (vals.x1_list[i], vals.y1_list[i])
                xold = vals.x1_list[i]
                yold = vals.y1_list[i]
            if vals.flags & 1:  # closed path
                path += ' z'
            attribs = {'d': path, 'style': style}
            etree.SubElement(layer, 'path', attribs)


def export_hatch(vals):
    # mandatory group codes : (10, 20, 70, 72, 92, 93) (x, y, fill, Edge Type, Path Type, Number of edges)
    if vals.has_x1 and vals.has_y1 and vals.has_fill and vals.has_edge_type \
            and vals.has_path_type and vals.has_num_edges:
        if len(vals.x1_list) > 1 and len(vals.y1_list) == len(vals.x1_list):
            # optional group codes : (11, 21, 40, 50, 51, 73) (x, y, r, angle1, angle2, CCW)
            i10 = 1  # count start points
            i11 = 0  # count line end points
            i40 = 0  # count circles
            i72 = 0  # count edge type flags
            path = ''
            for i in range(0, len(vals.num_edges_list)):
                xc = vals.x1_list[i10]
                yc = vals.y1_list[i10]
                if vals.edge_type_list[i72] == 2:  # arc
                    rm = scale * vals.radius_list[i40]
                    a1 = vals.angle_list[i40]
                    path += 'M %f,%f ' % (xc + rm * math.cos(a1 * math.pi / 180.0), yc + rm * math.sin(a1 * math.pi / 180.0))
                else:
                    a1 = 0
                    path += 'M %f,%f ' % (xc, yc)
                for j in range(0, vals.num_edges_list[i]):
                    if vals.path_type_list[i] & 2:  # polyline
                        if j > 0:
                            path += 'L %f,%f ' % (vals.x1_list[i10], vals.y1_list[i10])
                        if j == vals.path_type_list[i] - 1:
                            i72 += 1
                    elif vals.edge_type_list[i72] == 2:  # arc
                        xc = vals.x1_list[i10]
                        yc = vals.y1_list[i10]
                        rm = scale * vals.radius_list[i40]
                        a2 = vals.angle2_list[i40]
                        diff = (a2 - a1 + 360) % 360
                        sweep = 1 - vals.sweep_list[i40]  # sweep CCW
                        large = 0  # large-arc-flag
                        if diff:
                            path += 'A %f,%f 0.0 %d %d %f,%f ' % (rm, rm, large, sweep, xc + rm * math.cos(a2 * math.pi / 180.0), yc + rm * math.sin(a2 * math.pi / 180.0))
                        else:
                            path += 'A %f,%f 0.0 %d %d %f,%f ' % (rm, rm, large, sweep, xc + rm * math.cos((a1 + 180.0) * math.pi / 180.0), yc + rm * math.sin((a1 + 180.0) * math.pi / 180.0))
                            path += 'A %f,%f 0.0 %d %d %f,%f ' % (rm, rm, large, sweep, xc + rm * math.cos(a1 * math.pi / 180.0), yc + rm * math.sin(a1 * math.pi / 180.0))
                        i40 += 1
                        i72 += 1
                    elif vals.edge_type_list[i72] == 1:  # line
                        path += 'L %f,%f ' % (vals.x2_list[i11], vals.y2_list[i11])
                        i11 += 1
                        i72 += 1
                    i10 += 1
                path += "z "
            if vals.has_fill:
                style = formatStyle({'fill': '%s' % color})
            else:
                style = formatStyle({'fill': 'url(#Hatch)', 'fill-opacity': '1.0'})
            attribs = {'d': path, 'style': style}
            etree.SubElement(layer, 'path', attribs)


def export_dimension(vals):
    # mandatory group codes : (10, 11, 13, 14, 20, 21, 23, 24) (x1..4, y1..4)
    if vals.has_x1 and vals.has_x2 and vals.has_x3 and vals.has_x4 and \
            vals.has_y1 and vals.has_y2 and vals.has_y3 and vals.has_y4:
        dx = abs(vals.x1 - vals.x3)
        dy = abs(vals.y1 - vals.y3)
        if (vals.x1 == vals.x4) and dx > 0.00001:
            d = dx / scale
            dy = 0
            path = 'M %f,%f %f,%f' % (vals.x1, vals.y1, vals.x3, vals.y1)
        elif (vals.y1 == vals.y4) and dy > 0.00001:
            d = dy / scale
            dx = 0
            path = 'M %f,%f %f,%f' % (vals.x1, vals.y1, vals.x1, vals.y3)
        else:
            return
        attribs = {'d': path, 'style': style + '; marker-start: url(#DistanceX); marker-end: url(#DistanceX); stroke-width: 0.25px'}
        etree.SubElement(layer, 'path', attribs)
        x = vals.x2
        y = vals.y2
        size = 12  # default fontsize in px
        if vals.has_mtext:
            if vals.mtext in DIMTXT:
                size = scale * textscale * DIMTXT[vals.mtext]
                if size < 2:
                    size = 2
        attribs = {'x': '%f' % x, 'y': '%f' % y, 'style': 'font-size: %.3fpx; fill: %s; font-family: %s; text-anchor: middle; text-align: center' % (size, color, options.font)}
        if dx == 0:
            attribs.update({'transform': 'rotate (%f %f %f)' % (-90, x, y)})
        node = etree.SubElement(layer, 'text', attribs)
        tspan = node.add(inkex.Tspan())
        tspan.set('sodipodi:role', 'line')
        tspan.text = str(float('%.2f' % d))


def export_insert(vals):
    # mandatory group codes : (2, 10, 20) (block name, x, y)
    if vals.has_block_name and vals.has_x1 and vals.has_y1:
        x = vals.x1 + scale * xmin
        y = vals.y1 - scale * ymin - height
        elem = layer.add(inkex.Use())
        elem.set('xlink:href', '#' + quote(vals.block_name.replace(" ", "_").encode("utf-8")))
        elem.transform = 'translate(%f, %f)' % (x, y)
        if vals.has_insert_scale_x and vals.has_insert_scale_y:
            elem.transform.add_scale(vals.insert_scale_x, vals.insert_scale_y)


def export_block(vals):
    # mandatory group codes : (2) (block name)
    if vals.has_block_name:
        global block
        block = etree.SubElement(defs, 'symbol', {'id': vals.block_name.replace(" ", "_")})


def export_endblk(vals):
    global block
    block = defs  # initiallize with dummy


def export_attdef(vals):
    # mandatory group codes : (1, 2) (default, tag)
    if vals.has_default and vals.has_tag:
        vals.text_list.append(vals.tag)
        export_mtext(vals)


def generate_ellipse(xc, yc, xm, ym, w, a1, a2):
    rm = math.sqrt(xm * xm + ym * ym)
    a = math.atan2(ym, xm)
    diff = (a2 - a1 + 2 * math.pi) % (2 * math.pi)
    if abs(diff) > 0.0000001 and abs(diff - 2 * math.pi) > 0.0000001:  # open arc
        large = 0  # large-arc-flag
        if diff > math.pi:
            large = 1
        xt = rm * math.cos(a1)
        yt = w * rm * math.sin(a1)
        x1 = xt * math.cos(a) - yt * math.sin(a)
        y1 = xt * math.sin(a) + yt * math.cos(a)
        xt = rm * math.cos(a2)
        yt = w * rm * math.sin(a2)
        x2 = xt * math.cos(a) - yt * math.sin(a)
        y2 = xt * math.sin(a) + yt * math.cos(a)
        path = 'M %f,%f A %f,%f %f %d 0 %f,%f' % (xc + x1, yc - y1, rm, w * rm, -180.0 * a / math.pi, large, xc + x2, yc - y2)
    else:  # closed arc
        path = 'M %f,%f A %f,%f %f 1 0 %f,%f %f,%f %f 1 0 %f,%f z' % (xc + xm, yc - ym, rm, w * rm, -180.0 * a / math.pi, xc - xm, yc + ym, rm, w * rm, -180.0 * a / math.pi, xc + xm, yc - ym)
    attribs = {'d': path, 'style': style}
    etree.SubElement(layer, 'path', attribs)


def generate_gcodetools_point(xc, yc):
    elem = layer.add(inkex.PathElement())
    elem.style = 'stroke:none;fill:#ff0000'
    elem.set('inkscape:dxfpoint', '1')
    elem.path = 'm %s,%s 2.9375,-6.34375 0.8125,1.90625 6.84375,-6.84375 0,0 0.6875,0.6875 -6.84375,6.84375 1.90625,0.8125 z' % (xc, yc)


#   define DXF Entities and specify which Group Codes to monitor

class DxfInput(inkex.InputExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab", default="Options")
        pars.add_argument("--scalemethod", default="manual")
        pars.add_argument("--scale", default="1.0")
        pars.add_argument("--textscale", default="1.0")        
        pars.add_argument("--xmin", default="0.0")
        pars.add_argument("--ymin", default="0.0")
        pars.add_argument("--gcodetoolspoints", default=True, type=inkex.Boolean)
        pars.add_argument("--encoding", dest="input_encode", default="latin_1")
        pars.add_argument("--font", default="Arial")

    def load(self, stream):
        return stream

    def effect(self):
        global options
        global defs
        global entity
        global seqs
        global style
        global layer
        global scale
        global textscale
        global color
        global extrude
        global xmin
        global ymin
        global height
        global DIMTXT

        options = self.options

        doc = self.get_template(width=210 * 96 / 25.4, height=297 * 96 / 25.4)
        svg = doc.getroot()
        defs = svg.defs
        marker = etree.SubElement(defs, 'marker', {'id': 'DistanceX', 'orient': 'auto', 'refX': '0.0', 'refY': '0.0', 'style': 'overflow:visible'})
        etree.SubElement(marker, 'path', {'d': 'M 3,-3 L -3,3 M 0,-5 L  0,5', 'style': 'stroke:#000000; stroke-width:0.5'})
        pattern = etree.SubElement(defs, 'pattern', {'id': 'Hatch', 'patternUnits': 'userSpaceOnUse', 'width': '8', 'height': '8', 'x': '0', 'y': '0'})
        etree.SubElement(pattern, 'path', {'d': 'M8 4 l-4,4', 'stroke': '#000000', 'stroke-width': '0.25', 'linecap': 'square'})
        etree.SubElement(pattern, 'path', {'d': 'M6 2 l-4,4', 'stroke': '#000000', 'stroke-width': '0.25', 'linecap': 'square'})
        etree.SubElement(pattern, 'path', {'d': 'M4 0 l-4,4', 'stroke': '#000000', 'stroke-width': '0.25', 'linecap': 'square'})

        def _get_line():
            return self.document.readline().strip().decode(options.input_encode)

        def get_line():
            return _get_line(), _get_line()

        def get_group(group):
            line = get_line()
            if line[0] == group:
                return float(line[1])
            return 0.0

        xmax = xmin = ymin = 0.0
        height = 297.0 * 96.0 / 25.4  # default A4 height in pixels
        measurement = 0  # default inches
        line = get_line()
        polylines = 0
        flag = 0  # (0, 1, 2, 3) = (none, LAYER, LTYPE, DIMTXT)
        layer_colors = {}  # store colors by layer
        layer_nodes = {}  # store nodes by layer
        linetypes = {}  # store linetypes by name
        DIMTXT = {}  # store DIMENSION text sizes

        while line[0] and line[1] != 'BLOCKS':
            line = get_line()
            if options.scalemethod == 'file':
                if line[1] == '$MEASUREMENT':
                    measurement = get_group('70')
            elif options.scalemethod == 'auto':
                if line[1] == '$EXTMIN':
                    xmin = get_group('10')
                    ymin = get_group('20')
                if line[1] == '$EXTMAX':
                    xmax = get_group('10')
            if flag == 1 and line[0] == '2':
                layername = line[1]
                layer_nodes[layername] = svg.add(inkex.Layer.new(layername))
            if flag == 2 and line[0] == '2':
                linename = line[1]
                linetypes[linename] = []
            if flag == 3 and line[0] == '2':
                stylename = line[1]
            if line[0] == '2' and line[1] == 'LAYER':
                flag = 1
            if line[0] == '2' and line[1] == 'LTYPE':
                flag = 2
            if line[0] == '2' and line[1] == 'DIMSTYLE':
                flag = 3
            if flag == 1 and line[0] == '62':
                layer_colors[layername] = int(line[1])
            if flag == 2 and line[0] == '49':
                linetypes[linename].append(float(line[1]))
            if flag == 3 and line[0] == '140':
                DIMTXT[stylename] = float(line[1])
            if line[0] == '0' and line[1] == 'ENDTAB':
                flag = 0

        if options.scalemethod == 'file':
            scale = 25.4  # default inches
            if measurement == 1.0:
                scale = 1.0  # use mm
        elif options.scalemethod == 'auto':
            scale = 1.0
            if xmax > xmin:
                scale = 210.0 / (xmax - xmin)  # scale to A4 width
        else:
            scale = float(options.scale)  # manual scale factor
            xmin = float(options.xmin)
            ymin = float(options.ymin)
        bname = os.path.basename(options.input_file)
        svg.desc = f"{bname} - scale = {scale}, origin = ({xmin}, {ymin}), method = {options.scalemethod}"
        scale *= 96.0 / 25.4  # convert from mm to pixels
        textscale = float(options.textscale)

        if '0' not in layer_nodes:
            layer_nodes['0'] = svg.add(inkex.Layer.new('0'))

            layer_colors['0'] = 7

        for linename in linetypes.keys():  # scale the dashed lines
            linetype = ''
            for length in linetypes[linename]:
                if length == 0:  # test for dot
                    linetype += ' 0.5,'
                else:
                    linetype += '%.4f,' % math.fabs(length * scale)
            if linetype == '':
                linetypes[linename] = 'stroke-linecap: round'
            else:
                linetypes[linename] = 'stroke-dasharray:' + linetype

        entity = ''
        inENTITIES = False
        block = defs  # initiallize with dummy
        while line[0] and (line[1] != 'ENDSEC' or not inENTITIES):
            line = get_line()
            if line[1] == 'ENTITIES':
                inENTITIES = True
            if entity and vals.is_valid(line[0]):
                seqs.append(line[0])  # list of group codes
                if line[0] in ('1', '2', '3', '6', '8'):  # text value
                    val = line[1].replace(r'\~', ' ')
                    val = re.sub(r'\\A.*;', '', val)
                    val = re.sub(r'\\H.*;', '', val)
                    val = re.sub(r'\^I', '', val)
                    val = re.sub(r'{\\L', '', val)
                    val = re.sub(r'}', '', val)
                    val = re.sub(r'\\S.*;', '', val)
                    val = re.sub(r'\\W.*;', '', val)
                    val = val
                    val = re.sub(r'\\U\+([0-9A-Fa-f]{4})', re_hex2unichar, val)
                elif line[0] in ('62', '70', '92', '93'):
                    val = int(line[1])
                else:  # unscaled float value
                    val = float(line[1])
                vals[line[0]].append(val)
            elif has_export(line[1]):
                if has_export(entity):
                    if block != defs:  # in a BLOCK
                        layer = block
                    elif vals.has_layer_name:  # use Common Layer Name
                        if not vals.layer_name:
                            vals.layer_name = '0'  # use default name
                        if vals.layer_name not in layer_nodes:
                            layer_nodes[vals.layer_name] = svg.add(inkex.Layer.new(vals.layer_name))
                        layer = layer_nodes[vals.layer_name]
                    color = '#000000'  # default color
                    if vals.has_layer_name:
                        if vals.layer_name in layer_colors:
                            color = get_rgbcolor(layer_colors[vals.layer_name])
                    if vals.has_color:  # Common Color Number
                        color = get_rgbcolor(vals.color)
                    style = formatStyle({'stroke': '%s' % color, 'fill': 'none'})
                    w = 0.5  # default lineweight for POINT
                    if vals.has_line_weight:  # Common Lineweight
                        if vals.line_weight > 0:
                            w = 96.0 / 25.4 * vals.line_weight / 100.0
                            if w < 0.5:
                                w = 0.5
                            style = formatStyle({'stroke': '%s' % color, 'fill': 'none', 'stroke-width': '%.1f' % w})
                    if vals.has_line_type:  # Common Linetype
                        if vals.line_type in linetypes:
                            style += ';' + linetypes[vals.line_type]
                    extrude = 1.0
                    if vals.has_extrude:
                        extrude = float(vals.extrude)

                    vals.adjust_coords(xmin, ymin, scale, extrude, height)

                    if extrude == -1.0:  # reflect angles
                        if vals.has_angle and vals.has_angle2:
                            vals.angle2, vals.angle = 180.0 - vals.angle, 180.0 - vals.angle2
                    exporter = get_export(entity)
                    if exporter:
                        if entity == 'POINT':
                            exporter(vals, w)
                        else:
                            exporter(vals)

                if line[1] == 'POLYLINE':
                    inVertexs   = False
                    entity = 'LWPOLYLINE'
                    vals = ValueConstruct()
                    seqs = []
                    flag70  = 0     # default
                    while line[0] and (line[1] != 'SEQEND'):
                        line = get_line()
                        if line[1] == 'VERTEX' :
                            inVertexs   = True
                        if inVertexs == False :
                            if line[0]  == '6':  # 8:layer 6:line style
                                seqs.append(line[0])
                                vals[line[0]].append(line[1])
                            if line[0] =='70' : # flag
                                flg70 = int(line[1])
                        else :
                            if line[0] == '70' :
                                if int(line[1]) == 16 : # control point
                                    continue
                            if line[0] in ('10', '20') :  # vertexs
                                val = float(line[1])
                                seqs.append(line[0])
                                vals[line[0]].append(val)
                    seqs.append('70')
                    vals['70'].append(flag70)
                    continue
                
                entity = line[1]
                vals = ValueConstruct()
                seqs = []

        self.document = doc

def get_export(opt):
    return globals().get('export_' + opt.lower(), None)

def has_export(opt):
    return get_export(opt) is not None

if __name__ == '__main__':
    DxfInput().run()



#!/usr/bin/env python
"""
Optimise PNG file using optipng
"""

import os
import inkex
from inkex.extensions import TempDirMixin
from inkex.command import call

class PngOutput(TempDirMixin, inkex.RasterOutputExtension):
    def add_arguments(self, pars):
        pars.add_argument('--tab')
        # Lossless options
        pars.add_argument('--interlace', type=inkex.Boolean, default=True)
        pars.add_argument('--level', type=int, default=5)
        # Lossy options
        pars.add_argument('--bitdepth', type=inkex.Boolean, default=False)
        pars.add_argument('--color', type=inkex.Boolean, default=False)
        pars.add_argument('--palette', type=inkex.Boolean, default=False)

    def load(self, stream):
        """Load the PNG file (prepare it for optipng)"""
        self.png_file = os.path.join(self.tempdir, 'input.png')
        with open(self.png_file, 'wb') as fhl:
            fhl.write(stream.read())

    def save(self, stream):
        """Pass the PNG file to optipng with the options"""
        options = {
            'o': self.options.level,
            'i': int(self.options.interlace),
            'nb': not self.options.bitdepth,
            'nc': not self.options.color,
            'np': not self.options.palette,
        }
        call('optipng',
            self.png_file,
            oldie=True,
            clobber=True,
            **options)

        if os.path.isfile(self.png_file):
            with open(self.png_file, 'rb') as fhl:
                stream.write(fhl.read())
        

if __name__ == '__main__':
    PngOutput().run()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Martin Owens <doctormo@gmail.com>
#                    Thomas Holder <thomas.holder@schrodinger.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=arguments-differ
"""
Provide text based element classes interface.

Because text is not rendered at all, no information about a text's path
size or actual location can be generated yet.
"""

from ..paths import Path
from ..transforms import Transform, BoundingBox

from ._base import BaseElement, ShapeElement
from ._polygons import PathElementBase

class FlowRegion(ShapeElement):
    """SVG Flow Region (SVG 2.0)"""
    tag_name = 'flowRegion'

    def get_path(self):
        # This ignores flowRegionExcludes
        return sum([child.path for child in self], Path())

class FlowRoot(ShapeElement):
    """SVG Flow Root (SVG 2.0)"""
    tag_name = 'flowRoot'

    @property
    def region(self):
        """Return the first flowRegion in this flowRoot"""
        return self.findone('svg:flowRegion')

    def get_path(self):
        region = self.region
        return region.get_path() if region is not None else Path()

class FlowPara(ShapeElement):
    """SVG Flow Paragraph (SVG 2.0)"""
    tag_name = 'flowPara'

    def get_path(self):
        # XXX: These empty paths mean the bbox for text elements will be nothing.
        return Path()

class FlowDiv(ShapeElement):
    """SVG Flow Div (SVG 2.0)"""
    tag_name = 'flowDiv'

    def get_path(self):
        # XXX: These empty paths mean the bbox for text elements will be nothing.
        return Path()

class FlowSpan(ShapeElement):
    """SVG Flow Span (SVG 2.0)"""
    tag_name = 'flowSpan'

    def get_path(self):
        # XXX: These empty paths mean the bbox for text elements will be nothing.
        return Path()

class TextElement(ShapeElement):
    """A Text element"""
    tag_name = 'text'
    x = property(lambda self: self.uutounit(self.get('x', 0)))
    y = property(lambda self: self.uutounit(self.get('y', 0)))

    def get_path(self):
        return Path()

    def tspans(self):
        """Returns all children that are tspan elements"""
        return self.findall('svg:tspan')

    def get_text(self, sep="\n"):
        """Return the text content including tspans"""
        nodes = [self] + list(self.tspans())
        return sep.join([elem.text for elem in nodes if elem.text is not None])

    def shape_box(self, transform=None):
        """
        Returns a horrible bounding box that just contains the coord points
        of the text without width or height (which is impossible to calculate)
        """
        effective_transform = Transform(transform) * self.transform
        x, y = effective_transform.apply_to_point((self.x, self.y))
        bbox = BoundingBox(x, y)
        for tspan in self.tspans():
            bbox += tspan.bounding_box(effective_transform)
        return bbox

class TextPath(ShapeElement):
    """A textPath element"""
    tag_name = 'textPath'

    def get_path(self):
        return Path()

class Tspan(ShapeElement):
    """A tspan text element"""
    tag_name = 'tspan'
    x = property(lambda self: self.uutounit(self.get('x', 0)))
    y = property(lambda self: self.uutounit(self.get('y', 0)))

    @classmethod
    def superscript(cls, text):
        """Adds a superscript tspan element"""
        return cls(text, style="font-size:65%;baseline-shift:super")

    def get_path(self):
        return Path()

    def shape_box(self, transform=None):
        """
        Returns a horrible bounding box that just contains the coord points
        of the text without width or height (which is impossible to calculate)
        """
        effective_transform = Transform(transform) * self.transform
        x1, y1 = effective_transform.apply_to_point((self.x, self.y))
        fontsize = self.uutounit(self.style.get('font-size', '1em'))
        x2 = self.x + 0 # XXX This is impossible to calculate!
        y2 = self.y + float(fontsize)
        x2, y2 = effective_transform.apply_to_point((x2, y2))
        return BoundingBox((x1, x2), (y1, y2))


class SVGfont(BaseElement):
    """An svg font element"""
    tag_name = 'font'

class FontFace(BaseElement):
    """An svg font font-face element"""
    tag_name = 'font-face'

class Glyph(PathElementBase):
    """An svg font glyph element"""
    tag_name = 'glyph'

class MissingGlyph(BaseElement):
    """An svg font missing-glyph element"""
    tag_name = 'missing-glyph'



#!/usr/bin/env python
"""Make text lower case"""

import inkex

class Lowercase(inkex.TextExtension):
    """Convert to lowercase"""
    def process_chardata(self, text):
        return text.lower()

if __name__ == '__main__':
    Lowercase().run()



#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2013 <Public Domain>
#               2018 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from __future__ import print_function

import inkex
from inkex.localization import inkex_gettext as _

import hpgl_encoder

class HpglOutput(inkex.OutputExtension):
    """Save as HPGL Output"""
    def add_arguments(self, pars):
        pars.add_argument('--tab')
        pars.add_argument('--resolutionX', type=float, default=1016.0, help='Resolution X (dpi)')
        pars.add_argument('--resolutionY', type=float, default=1016.0, help='Resolution Y (dpi)')
        pars.add_argument('--pen', type=int, default=1, help='Pen number')
        pars.add_argument('--force', type=int, default=24, help='Pen force (g)')
        pars.add_argument('--speed', type=int, default=20, help='Pen speed (cm/s)')
        pars.add_argument('--orientation', default='90', help='Rotation (Clockwise)')
        pars.add_argument('--mirrorX', type=inkex.Boolean, default=False, help='Mirror X axis')
        pars.add_argument('--mirrorY', type=inkex.Boolean, default=False, help='Mirror Y axis')
        pars.add_argument('--center', type=inkex.Boolean, default=False, help='Center zero point')
        pars.add_argument('--overcut', type=float, default=1.0, help='Overcut (mm)')
        pars.add_argument('--precut', type=inkex.Boolean, default=True, help='Use precut')
        pars.add_argument('--flat', type=float, default=1.2, help='Curve flatness')
        pars.add_argument('--autoAlign', type=inkex.Boolean, default=True, help='Auto align')
        pars.add_argument('--toolOffset', type=float, default=0.25,\
            help='Tool (Knife) offset correction (mm)')

    def save(self, stream):
        self.options.debug = False
        # get hpgl data
        encoder = hpgl_encoder.hpglEncoder(self)
        try:
            hpgl = encoder.getHpgl()
        except hpgl_encoder.NoPathError:
            raise inkex.AbortExtension(
                _("No paths were found. Please convert objects you want into paths."))
        # convert raw HPGL to HPGL
        hpgl_init = 'IN'
       # if self.options.force > 0:
        #    hpgl_init += ';FS%d' % self.options.force
       # if self.options.speed > 0:
        #    hpgl_init += ';VS%d' % self.options.speed
        hpgl = hpgl_init + hpgl + ';SP0;PU0,0;IN; '
        stream.write(hpgl.encode('utf-8'))


if __name__ == '__main__':
    HpglOutput().run()


#!/usr/bin/env python
# coding=utf-8
#
# Copyright (C) 2010 Alvin Penner, penner@vaxxine.com
#
# - Voronoi Diagram algorithm and C code by Steven Fortune, 1987, http://ect.bell-labs.com/who/sjf/
# - Python translation to file voronoi.py by Bill Simons, 2005, http://www.oxfish.com/
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import random

import inkex
from inkex import PathElement, Pattern

import voronoi

def clip_line(x1, y1, x2, y2, w, h):
    if x1 < 0 and x2 < 0:
        return [0, 0, 0, 0]
    if x1 > w and x2 > w:
        return [0, 0, 0, 0]
    if x1 < 0:
        y1 = (y1 * x2 - y2 * x1) / (x2 - x1)
        x1 = 0
    if x2 < 0:
        y2 = (y1 * x2 - y2 * x1) / (x2 - x1)
        x2 = 0
    if x1 > w:
        y1 = y1 + (w - x1) * (y2 - y1) / (x2 - x1)
        x1 = w
    if x2 > w:
        y2 = y1 + (w - x1) * (y2 - y1) / (x2 - x1)
        x2 = w
    if y1 < 0 and y2 < 0:
        return [0, 0, 0, 0]
    if y1 > h and y2 > h:
        return [0, 0, 0, 0]
    if x1 == x2 and y1 == y2:
        return [0, 0, 0, 0]
    if y1 < 0:
        x1 = (x1 * y2 - x2 * y1) / (y2 - y1)
        y1 = 0
    if y2 < 0:
        x2 = (x1 * y2 - x2 * y1) / (y2 - y1)
        y2 = 0
    if y1 > h:
        x1 = x1 + (h - y1) * (x2 - x1) / (y2 - y1)
        y1 = h
    if y2 > h:
        x2 = x1 + (h - y1) * (x2 - x1) / (y2 - y1)
        y2 = h
    return [x1, y1, x2, y2]


class GenerateVoronoi(inkex.EffectExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--size", type=int, default=10, help="Average size of cell (px)")
        pars.add_argument("--border", type=int, default=0, help="Size of Border (px)")

    def effect(self):
        if not self.options.ids:
            return inkex.errormsg(_("Please select an object"))
        scale = self.svg.unittouu('1px')  # convert to document units
        self.options.size *= scale
        self.options.border *= scale
        obj = self.svg.selection.first()
        bbox = obj.bounding_box()
        mat = obj.composed_transform().matrix
        pattern = self.svg.defs.add(Pattern())
        pattern.set_random_id('Voronoi')
        pattern.set('width', str(bbox.width))
        pattern.set('height', str(bbox.height))
        pattern.set('patternUnits', 'userSpaceOnUse')
        pattern.patternTransform.add_translate(bbox.left - mat[0][2], bbox.top - mat[1][2])

        # generate random pattern of points
        c = voronoi.Context()
        pts = []
        b = float(self.options.border)  # width of border
        for i in range(int(bbox.width * bbox.height / self.options.size / self.options.size)):
            x = random.random() * bbox.width
            y = random.random() * bbox.height
            if b > 0:  # duplicate border area
                pts.append(voronoi.Site(x, y))
                if x < b:
                    pts.append(voronoi.Site(x + bbox.width, y))
                    if y < b:
                        pts.append(voronoi.Site(x + bbox.width, y + bbox.height))
                    if y > bbox.height - b:
                        pts.append(voronoi.Site(x + bbox.width, y - bbox.height))
                if x > bbox.width - b:
                    pts.append(voronoi.Site(x - bbox.width, y))
                    if y < b:
                        pts.append(voronoi.Site(x - bbox.width, y + bbox.height))
                    if y > bbox.height - b:
                        pts.append(voronoi.Site(x - bbox.width, y - bbox.height))
                if y < b:
                    pts.append(voronoi.Site(x, y + bbox.height))
                if y > bbox.height - b:
                    pts.append(voronoi.Site(x, y - bbox.height))
            elif x > -b and y > -b and x < bbox.width + b and y < bbox.height + b:
                pts.append(voronoi.Site(x, y))  # leave border area blank
            # dot = pattern.add(inkex.Rectangle())
            # dot.set('x', str(x-1))
            # dot.set('y', str(y-1))
            # dot.set('width', '2')
            # dot.set('height', '2')
        if len(pts) < 3:
            return inkex.errormsg("Please choose a larger object, or smaller cell size")

        # plot Voronoi diagram
        sl = voronoi.SiteList(pts)
        voronoi.voronoi(sl, c)
        path = ""
        for edge in c.edges:
            if edge[1] >= 0 and edge[2] >= 0:  # two vertices
                [x1, y1, x2, y2] = clip_line(c.vertices[edge[1]][0], c.vertices[edge[1]][1], c.vertices[edge[2]][0], c.vertices[edge[2]][1], bbox.width, bbox.height)
            elif edge[1] >= 0:  # only one vertex
                if c.lines[edge[0]][1] == 0:  # vertical line
                    xtemp = c.lines[edge[0]][2] / c.lines[edge[0]][0]
                    if c.vertices[edge[1]][1] > bbox.height / 2:
                        ytemp = bbox.height
                    else:
                        ytemp = 0
                else:
                    xtemp = bbox.width
                    ytemp = (c.lines[edge[0]][2] - bbox.width * c.lines[edge[0]][0]) / c.lines[edge[0]][1]
                [x1, y1, x2, y2] = clip_line(c.vertices[edge[1]][0], c.vertices[edge[1]][1], xtemp, ytemp, bbox.width, bbox.height)
            elif edge[2] >= 0:  # only one vertex
                if edge[0] >= len(c.lines):
                    xtemp = 0
                    ytemp = 0
                elif c.lines[edge[0]][1] == 0:  # vertical line
                    xtemp = c.lines[edge[0]][2] / c.lines[edge[0]][0]
                    if c.vertices[edge[2]][1] > bbox.height / 2:
                        ytemp = bbox.height
                    else:
                        ytemp = 0
                else:
                    xtemp = 0
                    ytemp = c.lines[edge[0]][2] / c.lines[edge[0]][1]
                [x1, y1, x2, y2] = clip_line(xtemp, ytemp, c.vertices[edge[2]][0], c.vertices[edge[2]][1], bbox.width, bbox.height)
            if x1 or x2 or y1 or y2:
                path += 'M %.3f,%.3f %.3f,%.3f ' % (x1, y1, x2, y2)

        patternstyle = {'stroke': '#000000', 'stroke-width': str(scale)}
        attribs = {'d': path, 'style': str(inkex.Style(patternstyle))}
        pattern.append(PathElement(**attribs))

        # link selected object to pattern
        style = {}
        if 'style' in obj.attrib:
            style = dict(inkex.Style.parse_str(obj.attrib['style']))
        style['fill'] = 'url(#%s)' % pattern.get('id')
        obj.attrib['style'] = str(inkex.Style(style))
        if isinstance(obj, inkex.Group):
            for node in obj:
                style = {}
                if 'style' in node.attrib:
                    style = dict(inkex.Style.parse_str(node.attrib['style']))
                style['fill'] = 'url(#%s)' % pattern.get('id')
                node.attrib['style'] = str(inkex.Style(style))

if __name__ == '__main__':
    GenerateVoronoi().run()



# coding=utf-8
#
# Copyright (C) 2018 Martin Owens <doctormo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Parsing inx files for checking and generating.
"""

import os
from inspect import isclass
from importlib import util
from lxml import etree

from .base import InkscapeExtension
from .utils import Boolean

NSS = {
    'inx': 'http://www.inkscape.org/namespace/inkscape/extension',
    'inkscape': 'http://www.inkscape.org/namespaces/inkscape',
}
SSN = dict([(b, a) for (a, b) in NSS.items()])

class InxLookup(etree.CustomElementClassLookup):
    """Custom inx xml file lookup"""
    def lookup(self, node_type, document, namespace, name):  # pylint: disable=unused-argument
        if name == 'param':
            return ParamElement
        return InxElement

INX_PARSER = etree.XMLParser()
INX_PARSER.set_element_class_lookup(InxLookup())

class InxFile:
    """Open an INX file and provide useful functions"""
    name = property(lambda self: self.xml._text('name'))
    ident = property(lambda self: self.xml._text('id'))
    slug = property(lambda self: self.ident.split('.')[-1].title().replace('_', ''))
    kind = property(lambda self: self.metadata['type'])
    warnings = property(lambda self: sorted(list(set(self.xml.warnings))))

    def __init__(self, filename):
        if isinstance(filename, str) and '<' in filename:
            filename = filename.encode('utf8')
        if isinstance(filename, bytes) and b'<' in filename:
            self.filename = None
            self.doc = etree.ElementTree(etree.fromstring(filename, parser=INX_PARSER))
        else:
            self.filename = os.path.basename(filename)
            self.doc = etree.parse(filename, parser=INX_PARSER)
        self.xml = self.doc.getroot()
        self.xml.warnings = []

    def __repr__(self):
        return f"<inx '{self.filename}' '{self.name}'>"

    @property
    def script(self):
        """Returns information about the called script"""
        command = self.xml.find_one('script/command')
        if command is None:
            return {}
        return {
            'interpreter': command.get('interpreter', None),
            'location': command.get('location', None),
            'script': command.text,
        }

    @property
    def extension_class(self):
        """Attempt to get the extension class"""
        script = self.script.get('script', None)
        if script is not None:
            name = script[:-3].replace('/', '.')
            spec = util.spec_from_file_location(name, script)
            mod = util.module_from_spec(spec)
            spec.loader.exec_module(mod)
            for value in mod.__dict__.values():
                if 'Base' not in name and isclass(value) and  value.__module__ == name \
                    and issubclass(value, InkscapeExtension):
                    return value
        return None

    @property
    def metadata(self):
        """Returns information about what type of extension this is"""
        effect = self.xml.find_one('effect')
        output = self.xml.find_one('output')
        inputs = self.xml.find_one('input')
        data = {}
        if effect is not None:
            template = self.xml.find_one('inkscape:templateinfo')
            if template is not None:
                data['type'] = 'template'
                data['desc'] = self.xml._text('templateinfo/shortdesc', nss='inkscape')
                data['author'] = self.xml._text('templateinfo/author', nss='inkscape')
            else:
                data['type'] = 'effect'
                data['preview'] = Boolean(effect.get('needs-live-preview', 'true'))
                data['objects'] = effect._text('object-type', 'all')
        elif inputs is not None:
            data['type'] = 'input'
            data['extension'] = inputs._text('extension')
            data['mimetype'] = inputs._text('mimetype')
            data['tooltip'] = inputs._text('filetypetooltip')
            data['name'] = inputs._text('filetypename')
        elif output is not None:
            data['type'] = 'output'
            data['dataloss'] = Boolean(output._text('dataloss', 'false'))
            data['extension'] = output._text('extension')
            data['mimetype'] = output._text('mimetype')
            data['tooltip'] = output._text('filetypetooltip')
            data['name'] = output._text('filetypename')
        return data

    @property
    def menu(self):
        """Return the menu this effect ends up in"""
        def _recurse_menu(parent):
            for child in parent.xpath('submenu'):
                yield child.get('name')
                for subchild in _recurse_menu(child):
                    yield subchild
                break # Not more than one menu chain?
        menu = self.xml.find_one('effect/effects-menu')
        return list(_recurse_menu(menu)) + [self.name]

    @property
    def params(self):
        """Get all params at all levels"""
        # Returns any params at any levels
        return list(self.xml.xpath('//param'))

class InxElement(etree.ElementBase):
    def set_warning(self, msg):
        root = self.get_root()
        if hasattr(root, 'warnings'):
            root.warnings.append(msg)

    def get_root(self):
        """Get the root document element from any element descendent"""
        if self.getparent() is not None:
            return self.getparent().get_root()
        return self

    def get_default_prefix(self):
        tag = self.get_root().tag
        if '}' in tag:
            (url, tag) = tag[1:].split('}', 1)
            return SSN.get(url, 'inx')
        self.set_warning("No inx xml prefix.")
        return None # no default prefix

    def apply_nss(self, xpath, nss=None):
        """Add prefixes to any xpath string"""
        if nss is None:
            nss = self.get_default_prefix()
        def _process(seg):
            if ':' in seg or not seg or not nss:
                return seg
            return f"{nss}:{seg}"
        return "/".join([_process(seg) for seg in xpath.split("/")])

    def xpath(self, xpath, nss=None):
        """Namespace specific xpath searches"""
        return super().xpath(self.apply_nss(xpath, nss=nss), namespaces=NSS)

    def find_one(self, name, nss=None):
        """Return the first element matching the given name"""
        for elem in self.xpath(name, nss=nss):
            return elem
        return None

    def _text(self, name, default=None, nss=None):
        """Get text content agnostically"""
        for pref in ('', '_'):
            elem = self.find_one(pref + name, nss=nss)
            if elem is not None and elem.text:
                if pref == '_':
                    self.set_warning(f"Use of old translation scheme: <_{name}...>")
                return elem.text
        return default

class ParamElement(InxElement):
    """
    A param in an inx file.
    """
    name = property(lambda self: self.get('name'))
    param_type = property(lambda self: self.get('type', 'string'))

    @property
    def options(self):
        """Return a list of option values"""
        if self.param_type == 'notebook':
            return [option.get('name')
                for option in self.xpath('page')]
        return [option.get('value')
            for option in self.xpath('option')]

    def __repr__(self):
        return f"<param name='{self.name}' type='{self.param_type}'>"
